"""Use cases for devices."""

import uuid
from collections.abc import AsyncIterator

from energy_assistant.db import AsyncSession
from energy_assistant.models.sessionlog import SessionLogEntry

from .schema import SessionLogEntrySchema


class ReadAllLogEntries:
    """Read all devices use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read all devices use case."""
        self.async_session = session

    async def execute(self, device_id: uuid.UUID | None) -> AsyncIterator[SessionLogEntrySchema]:
        """Execute the read all devices use case."""
        async with self.async_session() as session:
            if device_id is not None:
                async for entry in SessionLogEntry.read_by_device_id(session, device_id):
                    yield SessionLogEntrySchema(
                        start=entry.start,
                        end=entry.end,
                        text=entry.text,
                        device_id=entry.device_id,
                        solar_consumed_energy=entry.end_solar_consumed_energy - entry.start_solar_consumed_energy,
                        consumed_energy=entry.end_consumed_energy - entry.start_consumed_energy,
                    )
            else:
                async for entry in SessionLogEntry.read_all(session):
                    yield SessionLogEntrySchema(
                        start=entry.start,
                        end=entry.end,
                        text=entry.text,
                        device_id=entry.device_id,
                        solar_consumed_energy=entry.end_solar_consumed_energy - entry.start_solar_consumed_energy,
                        consumed_energy=entry.end_consumed_energy - entry.start_consumed_energy,
                    )



"""Drop date column

Revision ID: d43efed24fbc
Revises: 0bf7066b6d6c
Create Date: 2023-06-18 20:35:43.234174

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "d43efed24fbc"
down_revision = "0bf7066b6d6c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # op.drop_column("DeviceMeasurement", "grid_exported_energy")
    with op.batch_alter_table("DeviceMeasurement") as batch_op:
        batch_op.drop_column("date")


def downgrade() -> None:
    op.add_column("DeviceMeasurement", sa.Column("date", sa.DATE(), nullable=False))



"""Homeassistant importer."""

import logging
from datetime import datetime, timedelta
from math import isnan
from typing import Any

import pandas as pd
from sqlalchemy.ext.asyncio import AsyncSession

from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.devices import Location, State, StateId, StatesRepository
from energy_assistant.devices.home import Home
from energy_assistant.devices.homeassistant import HistoryState, Homeassistant
from energy_assistant.models.home import HomeMeasurement

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class StatesRepositoryWithHistory(StatesRepository):
    """A state repository for managing historical states."""

    def __init__(self, location: Location) -> None:
        """Create a StatesRepository instance."""
        self._read_states: pd.DataFrame = pd.DataFrame()
        self._template_states: dict[str, dict | Any] | None = None
        self._location = location
        self._selected_date: datetime | None = None

    def set_state_history(self, id: str, history: list[HistoryState], freq: pd.Timedelta) -> None:
        """Set the history of a state."""
        series = pd.Series({x.time_stamp: x.state for x in history})
        series.index = series.index.tz_convert(self._location.get_time_zone())  # type: ignore
        series = series.resample(freq).max().interpolate()
        self._read_states[id] = series

    def select_date(self, d: datetime) -> None:
        """Select the date for the next get_state call."""
        self._selected_date = d

    def get_state(self, id: StateId | str) -> State | None:
        """Get a state from the repository."""
        if self._selected_date is None:
            return None

        _id = id if isinstance(id, str) else id.id

        row = self._read_states.loc[self._selected_date]
        return State(_id, str(row[_id]))

    def get_numeric_states(self) -> dict[str, float]:
        """Get a states from the repository."""
        return {}

    def get_template_states(self) -> dict:
        """Get template states from the repository."""
        if self._template_states is None:
            self._template_states = {}
            states = self.get_numeric_states().items()
            for k, v in states:
                parts = k.split(".")
                if len(parts) > 1:
                    type = parts[0]
                    attribute = parts[1]
                    if type in self._template_states:
                        self._template_states[type][attribute] = v
                    else:
                        self._template_states[type] = {attribute: v}
                else:
                    self._template_states[k] = v
        return self._template_states

    def set_state(self, id: StateId, value: str, attributes: dict | None = None) -> None:
        """Set a state in the repository."""

    @property
    def channel(self) -> str:
        """Get the channel of the State Repository."""
        return "history"

    async def async_read_states(self) -> None:
        """Read the states from the channel asynchronously."""

    def read_states(self) -> None:
        """Read the states from the channel."""

    def write_states(self) -> None:
        """Write the states to the channel."""

    async def async_write_states(self) -> None:
        """Write the states to the channel."""


async def import_data(
    home: Home,
    hass: Homeassistant,
    session: AsyncSession,
    freq: pd.Timedelta,
    days_to_retrieve: int,
) -> None:
    """Import data from Homeassistant."""
    states_repository = StatesRepositoryWithHistory(await hass.get_location())

    start_date = datetime.now(tz=await hass.get_timezone()).replace(
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
    ) - timedelta(days=days_to_retrieve)

    variables = home.get_variables()
    for variable in variables:
        try:
            values = await hass.get_history(variable, start_time=start_date)
            states_repository.set_state_history(variable, values, freq)
        except Exception:
            LOGGER.exception(f"error during fetching the history of state {variable} for Home Assistant")

    # states_repository._read_states.to_csv(Path(settings.DATA_FOLDER)/"state_history.csv")
    energy_state = home.create_home_energy_state_clone()
    for row in states_repository._read_states.iterrows():
        try:
            d: datetime = row[0]  # type: ignore
            states_repository.select_date(d)
            await energy_state.async_update_state(states_repository)
            home_measurement = await HomeMeasurement.read_by_date(session, d.date(), False)
            if home_measurement is None:
                home_measurement = await HomeMeasurement.create(
                    session,
                    name=home.name,
                    measurement_date=d.date(),
                    solar_produced_energy=energy_state.produced_solar_energy,
                    grid_imported_energy=energy_state.grid_imported_energy,
                    grid_exported_energy=energy_state.grid_exported_energy,
                    device_measurements=[],
                )
            elif not isnan(energy_state.grid_exported_energy):
                await home_measurement.update(
                    session,
                    name=home.name,
                    measurement_date=d.date(),
                    solar_produced_energy=energy_state.produced_solar_energy,
                    grid_imported_energy=energy_state.grid_imported_energy,
                    grid_exported_energy=energy_state.grid_exported_energy,
                )
            else:
                LOGGER.error("invalid grid exported value")
        except Exception:
            LOGGER.exception("error during fetching data from the database")
    await session.flush()
    await session.commit()



"""Schemas for all the models."""

import datetime
import uuid

from pydantic import BaseModel, ConfigDict


class DeviceMeasurementSchema(BaseModel):
    """Schema class for a device measurement."""

    id: int
    # name: str
    solar_consumed_energy: float
    consumed_energy: float
    home_measurement_id: int
    device_id: uuid.UUID
    measurement_date: datetime.date | None

    model_config = ConfigDict(from_attributes=True)


class DeviceSchema(BaseModel):
    """Schema class for a device."""

    id: uuid.UUID
    name: str
    icon: str
    type: str
    config: str
    supported_power_modes: list[str] | None = None
    power_mode: str | None = None

    model_config = ConfigDict(from_attributes=True)


class HomeMeasurementSchema(BaseModel):
    """Schema class for a home measurement."""

    id: int
    name: str
    solar_consumed_energy: float
    consumed_energy: float
    solar_produced_energy: float
    grid_imported_energy: float
    grid_exported_energy: float
    measurement_date: datetime.date

    device_measurements: list[DeviceMeasurementSchema]

    model_config = ConfigDict(from_attributes=True)



"""Device type registry for device data."""

import logging
from dataclasses import dataclass
from pathlib import Path

import yaml

from energy_assistant.constants import (
    DEFAULT_NOMINAL_DURATION,
    DEFAULT_NOMINAL_POWER,
    ROOT_LOGGER_NAME,
)

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


@dataclass(frozen=True, eq=True)
class DeviceTypeId:
    """The of a device type."""

    manufacturer: str
    model: str


@dataclass
class DeviceType:
    """The device type of a device."""

    icon: str
    nominal_power: float
    nominal_duration: float
    constant: bool
    state_on_threshold: float
    state_off_threshold: float
    state_off_upper: float
    state_off_lower: float
    state_off_for: float
    trailing_zeros_for: float


class DeviceTypeRegistry:
    """The registry for device types."""

    def __init__(self) -> None:
        """Create a Device Type Registry instance."""
        self._registry: dict[DeviceTypeId, DeviceType] = {}

    def load(self, config_folder: Path) -> None:
        """Load the registry from the configuration folder."""
        for config_file in config_folder.glob("**/*/*.yaml"):
            self.load_device_type_file(config_file)

    def load_device_type_file(self, filename: Path) -> None:
        """Load a device type config file and add it to the registry."""
        with filename.open() as stream:
            try:
                config = yaml.safe_load(stream)
            except yaml.YAMLError:
                LOGGER.exception("Yaml error while parsing device type file")
            except Exception:
                LOGGER.exception("error while parsing device type file")
            else:
                device_type_config = config.get("device_type")
                if device_type_config is None:
                    LOGGER.error(f"Device type config file {filename} does not contain a device_type item.")
                else:
                    manufacturer = device_type_config.get("manufacturer")
                    model = device_type_config.get("model")
                    icon = device_type_config.get("icon")
                    nominal_power = device_type_config.get("nominal_power", DEFAULT_NOMINAL_POWER)
                    nominal_duration = device_type_config.get("nominal_duration", DEFAULT_NOMINAL_DURATION)
                    constant = device_type_config.get("constant", False)
                    if model is None or manufacturer is None or icon is None:
                        LOGGER.error(f"Manufacturer or Model or Icon not set in device type config file {filename}")
                    else:
                        device_state_config = device_type_config.get("state")
                        state_on_threshold: float | None = None
                        state_on_config = device_state_config.get("state_on")
                        if state_on_config is not None:
                            state_on_threshold = state_on_config.get("threshold")

                        state_off_threshold: float = 0.0
                        state_off_upper: float | None = None
                        state_off_lower: float | None = None
                        state_off_for: float | None = None
                        state_off_config = device_state_config.get("state_off")
                        if state_off_config is not None:
                            state_off_threshold = state_off_config.get("threshold", 0.0)
                            state_off_upper = state_off_config.get("upper")
                            state_off_lower = state_off_config.get("lower")
                            state_off_for = state_off_config.get("for")
                            trailing_zeros_for = state_off_config.get("trailing_zeros_for")
                        if (
                            state_on_threshold is not None
                            and state_off_upper is not None
                            and state_off_lower is not None
                            and state_off_for is not None
                            and trailing_zeros_for is not None
                        ):
                            device_type = DeviceType(
                                icon=icon,
                                nominal_power=nominal_power,
                                nominal_duration=nominal_duration,
                                constant=constant,
                                state_on_threshold=state_on_threshold,
                                state_off_threshold=state_off_threshold,
                                state_off_upper=state_off_upper,
                                state_off_lower=state_off_lower,
                                state_off_for=state_off_for,
                                trailing_zeros_for=trailing_zeros_for,
                            )
                            self._registry[DeviceTypeId(manufacturer=manufacturer, model=model)] = device_type

    def get_device_type(self, manufacturer: str, model: str) -> DeviceType | None:
        """Get the device type for a given manufacturer and model."""
        return self._registry.get(DeviceTypeId(manufacturer=manufacturer, model=model))



"""Schemas for home measurement api."""

from pydantic import BaseModel

from energy_assistant.models.schema import HomeMeasurementSchema


class ReadHomeMeasurementResponse(HomeMeasurementSchema):
    """API Response for reading home measurements."""


class ReadAllHomeMeasurementResponse(BaseModel):
    """API Response for reading home measurements."""

    home_measurements: list[HomeMeasurementSchema]



"""Heat pump device implementation."""

import uuid

from energy_assistant import Optimizer
from energy_assistant.constants import POWER_HYSTERESIS
from energy_assistant.devices.device import DeviceWithState
from energy_assistant.devices.state_value import StateValue

from . import (
    LoadInfo,
    OnOffState,
    PowerModes,
    SessionStorage,
    State,
    StateId,
    StatesRepository,
)
from .analysis import DataBuffer
from .config import DeviceConfigMissingParameterError, get_config_param
from .homeassistant import HOMEASSISTANT_CHANNEL, assign_if_available

DEFAULT_NOMINAL_POWER = 5000


def numeric_value(value: str | None) -> float | None:
    """Convert into a number."""
    if value is None:
        return None
    try:
        return float(value)
    except ValueError:
        return None


class HeatPumpDevice(DeviceWithState):
    """Stiebel Eltron heatpump. This can be either a water heating part or a heating part."""

    def __init__(self, device_id: uuid.UUID, session_storage: SessionStorage) -> None:
        """Create a Stiebel Eltron heatpump."""
        super().__init__(device_id, session_storage)
        self._actual_temp: State | None = None
        self._actual_temp_entity_id: str = ""
        self._comfort_target_temperature_entity_id: str | None = None
        self._state_entity_id: str = ""
        self._state: State | None = None
        self._consumed_energy: State | None = None
        self._icon = "mdi-heat-pump"
        self._nominal_power: float = DEFAULT_NOMINAL_POWER
        self._target_temperature_normal: float | None = None
        self._target_temperature_pv: float | None = None
        self._consumed_energy_value: StateValue | None = None

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        super().configure(config)
        energy_config = config.get("energy")
        if energy_config is not None:
            self._consumed_energy_value = StateValue(energy_config)
        else:
            msg = "energy"
            raise DeviceConfigMissingParameterError(msg)
        self._actual_temp_entity_id = get_config_param(config, "temperature")
        self._nominal_power = config.get("nominal_power", DEFAULT_NOMINAL_POWER)
        self._comfort_target_temperature_entity_id = config.get("comfort_target_temperature")
        self._target_temperature_normal = numeric_value(config.get("target_temperature_normal"))
        self._target_temperature_pv = numeric_value(config.get("target_temperatrure_pv"))
        if (
            self._target_temperature_normal is not None
            and self._target_temperature_pv is not None
            and self._comfort_target_temperature_entity_id is not None
        ):
            self._supported_power_modes.add(PowerModes.PV)
            self._supported_power_modes.add(PowerModes.OPTIMIZED)

        self._state_entity_id = get_config_param(config, "state")

    @property
    def type(self) -> str:
        """The device type."""
        return "heat-pump"

    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the state of the Stiebel Eltron device."""
        old_state = self.state == "on"
        self._state = assign_if_available(self._state, state_repository.get_state(self._state_entity_id))
        new_state = self.state == "on"

        self._consumed_energy = assign_if_available(
            self._consumed_energy,
            self._consumed_energy_value.evaluate(state_repository) if self._consumed_energy_value is not None else None,
        )
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        self._actual_temp = assign_if_available(
            self._actual_temp,
            state_repository.get_state(self._actual_temp_entity_id),
        )

        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        await self.update_session(old_state, new_state, "Water heater")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if (
            self._target_temperature_normal is not None
            and self._target_temperature_pv is not None
            and self._comfort_target_temperature_entity_id is not None
        ):
            current_target_temperature = state_repository.get_state(self._comfort_target_temperature_entity_id)
            if current_target_temperature is not None:
                target_temperature: float = current_target_temperature.numeric_value
                if self.power_mode == PowerModes.PV:
                    if self.state == "off":
                        avg_300 = grid_exported_power_data.get_average_for(300)
                        if avg_300 > self.requested_additional_power * (1 + POWER_HYSTERESIS):
                            target_temperature = self._target_temperature_pv
                        elif avg_300 < self.requested_additional_power * (1 - POWER_HYSTERESIS):
                            target_temperature = self._target_temperature_normal
                elif self.power_mode == PowerModes.OPTIMIZED:
                    target_temperature = self._get_temperature_for_optimized(optimizer, target_temperature)
                if target_temperature != current_target_temperature.numeric_value:
                    state_repository.set_state(
                        StateId(
                            id=self._comfort_target_temperature_entity_id,
                            channel=HOMEASSISTANT_CHANNEL,
                        ),
                        str(target_temperature),
                    )

    def _get_temperature_for_optimized(self, optimizer: Optimizer, current_target_temperature: float) -> float:
        if (
            self.state == "off"
            and self._target_temperature_normal is not None
            and self._target_temperature_pv is not None
        ):
            power = optimizer.get_optimized_power(self._id)
            if power > 0:
                return self._target_temperature_pv
            return self._target_temperature_normal
        return current_target_temperature

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if self.power_mode == PowerModes.OPTIMIZED:
            return LoadInfo(
                device_id=self.id,
                nominal_power=self._nominal_power,
                duration=1800,  # 0.5h -> TODO: make this configurable
                is_continous=False,
                is_deferrable=True,
            )
        return None

    @property
    def requested_additional_power(self) -> float:
        """How much power the device could consume in pv mode."""
        # TODO: Consider the actual temperature
        return self._nominal_power if self.state == "off" else 0.0

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        return self._consumed_energy.numeric_value if self._consumed_energy else 0.0

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is heating."""
        return self._state.value if self._state and self._state.value else "unknown"

    @property
    def power(self) -> float:
        """Current power consumption of the device."""
        if self._state is not None:
            return self._nominal_power if self._state.value == "on" else 0.0
        return 0.0

    @property
    def actual_temperature(self) -> float:
        """The actual temperature of the heating or water."""
        return self._actual_temp.numeric_value if self._actual_temp else 0.0

    @property
    def icon(self) -> str:
        """Icon of the device."""
        return self._icon

    @property
    def available(self) -> bool:
        """Is the device available?."""
        return (
            self._consumed_energy is not None
            and self._consumed_energy.available
            and self._actual_temp is not None
            and self._actual_temp.available
            and self._state is not None
            and self._state.available
        )

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore the previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = State("", str(consumed_energy))

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        result: dict[str, str] = {
            **super().attributes,
            "actual_temperature": f"{self.actual_temperature} °C",
        }
        return result


class SubHeatPump(DeviceWithState):
    """The heating or the water heating part of the heat pump."""

    def __init__(self, device_id: uuid.UUID, session_storage: SessionStorage, name: str) -> None:
        """Create a sub heat pump instance."""
        super().__init__(device_id, session_storage)

        self._actual_temp_entity_id: str = ""
        self._actual_temp: State | None = None
        self._state: State | None = None
        self._state_entity_id: str = ""
        self._consumed_energy: State | None = None
        self._consumed_energy_value: StateValue | None = None

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        sub_device_config = {**config, "name": self.name, "id": uuid.uuid4()}
        super().configure(sub_device_config)
        energy_config = config.get("energy")
        if energy_config is not None:
            self._consumed_energy_value = StateValue(energy_config)
        else:
            msg = "energy"
            raise DeviceConfigMissingParameterError(msg)
        self._actual_temp_entity_id = get_config_param(config, "temperature")
        self._state_entity_id = get_config_param(config, "state")

    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the state of the SGReady Heatpump."""

        # old_state = self.state == OnOffState.ON
        self._state = assign_if_available(self._state, state_repository.get_state(self._state_entity_id))
        # new_state = self.state == OnOffState.ON
        self._consumed_energy = assign_if_available(
            self._consumed_energy,
            self._consumed_energy_value.evaluate(state_repository) if self._consumed_energy_value is not None else None,
        )
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        self._actual_temp = assign_if_available(
            self._actual_temp,
            state_repository.get_state(self._actual_temp_entity_id),
        )

        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        # TODO: implement session tracking for sub devices
        # await self.update_session(old_state, new_state, self._name)

    @property
    def actual_temperature(self) -> float:
        """The actual temperature of the heating or water."""
        return self._actual_temp.numeric_value if self._actual_temp else 0.0

    @property
    def state(self) -> OnOffState:
        """The state of the device. The state is `on` in case the device is heating."""
        return OnOffState.from_str(self._state.value) if self._state and self._state.value else OnOffState.UNKNOWN

    @property
    def available(self) -> bool:
        """Is the device available?."""
        return (
            self._consumed_energy is not None
            and self._consumed_energy.available
            and self._actual_temp is not None
            and self._actual_temp.available
            and self._state is not None
            and self._state.available
        )

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        return self._consumed_energy.numeric_value if self._consumed_energy else 0.0

    @property
    def power(self) -> float:
        """Current power consumption of the device."""
        raise NotImplementedError

    @property
    def type(self) -> str:
        """The device type."""
        return "sg-ready-heat-pump"

    @property
    def icon(self) -> str:
        """The icon of the device."""
        return "mdi-heat-pump"

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        raise NotImplementedError

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        result: dict[str, str] = {
            **super().attributes,
            "actual_temperature": f"{self.actual_temperature} °C",
        }
        return result


class SGReadyHeatPumpDevice(DeviceWithState):
    """SG Ready heatpump, supporting water and heating temperature."""

    def __init__(self, device_id: uuid.UUID, session_storage: SessionStorage) -> None:
        """Create a SG Ready heatpump."""
        super().__init__(device_id, session_storage)

        self._heating = SubHeatPump(device_id, session_storage, "heating")

        self._water_heating = SubHeatPump(device_id, session_storage, "water_heating")

        self._nominal_power: float = DEFAULT_NOMINAL_POWER
        self._sg_ready_switch_entity_id: str | None = None

        self._icon = "mdi-heat-pump"

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        super().configure(config)
        if self._heating is not None:
            self._heating.configure(config.get("heating", {}))
        if self._water_heating is not None:
            self._water_heating.configure(config.get("water", {}))
        self._nominal_power = config.get("nominal_power", DEFAULT_NOMINAL_POWER)
        self._sg_ready_switch_entity_id = config.get("sg_ready")
        if self._sg_ready_switch_entity_id is not None:
            self._supported_power_modes.add(PowerModes.PV)
            self._supported_power_modes.add(PowerModes.OPTIMIZED)

    @property
    def type(self) -> str:
        """The device type."""
        return "sg-ready-heat-pump"

    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the state of the SGReady heatpump."""

        old_state = self.state == OnOffState.ON

        if self._heating is not None:
            await self._heating.update_state(state_repository, self_sufficiency)

        if self._water_heating is not None:
            await self._water_heating.update_state(state_repository, self_sufficiency)

        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)

        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        await self.update_session(old_state, self.state == OnOffState.ON, "Heat pump")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if self._sg_ready_switch_entity_id is not None:
            current_sg_ready_state = state_repository.get_state(self._sg_ready_switch_entity_id)
            if current_sg_ready_state is not None:
                sg_ready_state = OnOffState.from_str(current_sg_ready_state.value)
                if self.power_mode == PowerModes.PV:
                    if self.state == OnOffState.OFF:
                        avg_300 = grid_exported_power_data.get_average_for(300)
                        if avg_300 > self.requested_additional_power * (1 + POWER_HYSTERESIS):
                            sg_ready_state = OnOffState.ON
                        elif avg_300 < self.requested_additional_power * (1 - POWER_HYSTERESIS):
                            sg_ready_state = OnOffState.OFF
                elif self.power_mode == PowerModes.OPTIMIZED:
                    sg_ready_state = self._get_state_for_optimized(
                        optimizer,
                        OnOffState.from_str(current_sg_ready_state.value),
                    )
                if sg_ready_state != OnOffState.from_str(current_sg_ready_state.value):
                    state_repository.set_state(
                        StateId(
                            id=self._sg_ready_switch_entity_id,
                            channel=HOMEASSISTANT_CHANNEL,
                        ),
                        str(sg_ready_state),
                    )

    @property
    def state(self) -> OnOffState:
        """State of the device."""
        if self._heating is not None and self._heating.state == OnOffState.ON:
            return OnOffState.ON
        if self._water_heating is not None and self._water_heating.state == OnOffState.ON:
            return OnOffState.ON
        return OnOffState.OFF

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if self.power_mode == PowerModes.OPTIMIZED:
            return LoadInfo(
                device_id=self.id,
                nominal_power=self._nominal_power,
                duration=1800,  # 0.5h -> TODO: make this configurable
                is_continous=False,
                is_deferrable=True,
            )
        return None

    def _get_state_for_optimized(self, optimizer: Optimizer, current_state: OnOffState) -> OnOffState:
        if self.state == OnOffState.OFF:
            power = optimizer.get_optimized_power(self._id)
            if power > 0:
                return OnOffState.ON
            return OnOffState.ON
        return current_state

    @property
    def requested_additional_power(self) -> float:
        """How much power the device could consume in pv mode."""
        # TODO: Consider the actual temperature
        return self._nominal_power if self.state == "off" else 0.0

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        result: float = 0.0
        if self._heating is not None:
            result += self._heating.consumed_energy
        if self._water_heating is not None:
            result += self._water_heating.consumed_energy
        return result

    @property
    def consumed_solar_energy(self) -> float:
        """Consumed solar energy in kWh."""
        result: float = 0.0
        if self._heating is not None:
            result += self._heating.consumed_solar_energy
        if self._water_heating is not None:
            result += self._water_heating.consumed_solar_energy
        return result

    @property
    def power(self) -> float:
        """Current power consumption of the device."""
        return self._nominal_power if self.state == OnOffState.ON else 0.0

    @property
    def icon(self) -> str:
        """Icon of the device."""
        return self._icon

    @property
    def available(self) -> bool:
        """Is the device available?."""
        if self._heating is not None and self._heating.available:
            return True
        return self._water_heating is not None and self._water_heating.available

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore the previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        heating_attributes = (
            {"heating_" + k: v for k, v in self._heating.attributes.items()} if self._heating is not None else {}
        )
        del heating_attributes["heating_state"]
        water_heating_attributes = (
            {"water_heating_" + k: v for k, v in self._water_heating.attributes.items()}
            if self._water_heating is not None
            else {}
        )
        del water_heating_attributes["water_heating_state"]

        heatpump_state = str(OnOffState.OFF)
        if self._heating is not None and self._heating.state == OnOffState.ON:
            heatpump_state = "Heating"
        elif self._water_heating is not None and self._water_heating.state == OnOffState.ON:
            heatpump_state = "Water heating"
        return {
            **super().attributes,
            **heating_attributes,
            **water_heating_attributes,
            "heatpump_state": heatpump_state,
        }



"""Interface to the homeassistant instance."""

import asyncio
import logging
import math
import uuid
from dataclasses import dataclass
from datetime import UTC, datetime, tzinfo
from enum import StrEnum
from zoneinfo import ZoneInfo

import pandas as pd
import requests  # type: ignore
from aiohttp import ClientResponse, ClientSession, TCPConnector
from hass_client import HomeAssistantClient  # type: ignore
from hass_client.exceptions import BaseHassClientError  # type: ignore
from hass_client.utils import get_websocket_url  # type: ignore

from energy_assistant import Optimizer
from energy_assistant.constants import (
    DEFAULT_NOMINAL_DURATION,
    DEFAULT_NOMINAL_POWER,
    POWER_HYSTERESIS,
    ROOT_LOGGER_NAME,
)
from energy_assistant.devices.analysis import DataBuffer
from energy_assistant.devices.registry import DeviceType, DeviceTypeRegistry
from energy_assistant.devices.state_value import StateValue

from . import (
    LoadInfo,
    Location,
    OnOffState,
    PowerModes,
    SessionStorage,
    State,
    StateId,
    StatesRepository,
    StatesSingleRepository,
    assign_if_available,
)
from .config import DeviceConfigMissingParameterError, get_config_param
from .device import DeviceWithState

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)
UNAVAILABLE = "unavailable"

HOMEASSISTANT_CHANNEL = "ha"


class HomeassistantState(State):
    """Abstract base class for states."""

    def __init__(self, id: str, value: str, attributes: dict | None = None) -> None:
        """Create a State instance."""
        super().__init__(id, value, attributes)

        if value == UNAVAILABLE:
            self._available = False
        else:
            self._available = True

    @property
    def name(self) -> str:
        """The name of the State."""
        if self._attributes is not None:
            return str(self._attributes.get("friendly_name"))
        return self._id

    @property
    def unit(self) -> str:
        """Unit of the state."""
        if self._attributes is not None:
            return str(self._attributes.get("unit_of_measurement"))
        return ""


@dataclass
class HistoryState:
    """Represents a history of a state."""

    time_stamp: datetime
    state: float


def convert_statistics(value: dict) -> dict:
    """Convert the times in a Home Assistant dict to date time values."""
    result = value.copy()
    result["start"] = datetime.fromtimestamp(value["start"] / 1000, UTC)
    result["end"] = datetime.fromtimestamp(value["end"] / 1000, UTC)
    return result


def convert_float(value: str) -> float:
    """Convert a value from Home Assistant to a float."""
    if value == "unavailable":
        return math.nan
    return float(value)


def convert_history(value: dict) -> HistoryState:
    """Convert a history state to a HistoryState instance."""
    return HistoryState(time_stamp=datetime.fromtimestamp(value["lu"], UTC), state=convert_float(value["s"]))


class StatisticsType(StrEnum):
    """The type of statistics which can be requested from Home Assistant."""

    MAX = "max"
    MEAN = "mean"
    MIN = "min"
    CHANGE = "change"


class StatisticsPeriod(StrEnum):
    """The statistics period which can be requested from Home Assistant."""

    FIVE_Min = "5minute"
    DAY = "day"
    HOUR = "hour"
    WEEK = "week"
    MONTH = "month"


class SourceType(StrEnum):
    """The type of the energy source."""

    GRID = "grid"
    SOLAR = "solar"


@dataclass
class EnergySource:
    """An energy source with the corresponding energy sensors."""

    source_type: SourceType
    flow_from: str
    flow_to: str | None


class HomeAssistantCommunicationError(Exception):
    """Home Assistant Communication Error."""

    def __init__(self, response: ClientResponse) -> None:
        """Create a HomeAssistantCommunicationError instance."""
        super().__init__(f"Error during communication with Home Assistant. Url={response.url} Status={response.status}")


class Homeassistant(StatesSingleRepository):
    """Home assistant proxy."""

    session: ClientSession
    hass: HomeAssistantClient
    _listen_task: asyncio.Task | None = None

    _time_zone: tzinfo | None

    def __init__(self, url: str, token: str, demo_mode: bool) -> None:
        """Create an instance of the Homeassistant class."""
        super().__init__(HOMEASSISTANT_CHANNEL)
        self._url = url
        self._token = token
        self._demo_mode = demo_mode is not None and demo_mode
        self._time_zone: tzinfo | None = None

    async def connect(self) -> None:
        """Connect to the homeassistant instance."""
        loop = asyncio.get_running_loop()
        self.session = ClientSession(
            loop=loop,
            connector=TCPConnector(
                ssl=False,
                enable_cleanup_closed=True,
                limit=4096,
                limit_per_host=100,
            ),
        )
        url = get_websocket_url(self._url)

        self.hass = HomeAssistantClient(url, self._token, self.session)
        await self.hass.connect()
        self._listen_task = asyncio.create_task(self._hass_listener())
        LOGGER.info("Connected to Homeassistant version %s", self.hass.version)

    async def disconnect(self) -> None:
        """Disconnects from home assistant."""
        if self._listen_task and not self._listen_task.done():
            self._listen_task.cancel()
        await self.hass.disconnect()
        await self.session.close()

    async def _hass_listener(self) -> None:
        """Start listening on the HA websockets."""
        if self._listen_task is not None:
            while not self._listen_task.cancelled():
                try:
                    # start listening will block until the connection is lost/closed
                    await self.hass.start_listening()
                except BaseHassClientError as err:
                    LOGGER.warning("Connection to HA lost due to error: %s", err)
                LOGGER.info("Connection to HA lost. Reconnecting.")
                # schedule a reload of the provider
                # self.mass.call_later(5, self.mass.config.reload_provider(self.instance_id))

    @property
    def url(self) -> str:
        """URL of the home assistant instance."""
        return self._url

    @property
    def token(self) -> str:
        """Token of the home assistant instance."""
        return self._token

    async def get_statistics(
        self,
        entity_id: str,
        start_time: datetime | None = None,
        types: list[StatisticsType] | None = None,
        period: StatisticsPeriod = StatisticsPeriod.HOUR,
    ) -> list[dict]:
        """Read the statistics for an entity."""
        if start_time is None:
            start_time = datetime.now(tz=await self.get_timezone()).replace(hour=0, minute=0, second=0, microsecond=0)
        statistics = await self.hass.send_command(
            "recorder/statistics_during_period",
            period=period,
            start_time=start_time.isoformat(),
            statistic_ids=[entity_id],
            types=types if types is not None else [],
        )
        if entity_id in statistics:
            return [convert_statistics(value) for value in statistics[entity_id]]
        return []

    async def get_history(
        self,
        entity_id: str,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
    ) -> list[HistoryState]:
        """Get the history of a state from Home Assistant."""
        if start_time is None:
            start_time = datetime.now(tz=await self.get_timezone()).replace(hour=0, minute=0, second=0, microsecond=0)
        if end_time is None:
            end_time = datetime.now(tz=await self.get_timezone())
        history = await self.hass.send_command(
            "history/history_during_period",
            start_time=start_time.isoformat(),
            end_time=end_time.isoformat(),
            entity_ids=[entity_id],
            no_attributes=True,
            minimal_response=True,
        )
        if entity_id in history:
            return [convert_history(value) for value in history[entity_id]]
        return []

    async def get_energy_info(self) -> dict:
        """Get the energy info from Home Assistant."""
        return await self.hass.send_command("energy/info")

    async def get_energy_prefs(self) -> list[EnergySource]:
        """Get the energy configuration from Home Assistant."""
        prefs = await self.hass.send_command("energy/get_prefs")
        sources = prefs.get("energy_sources")
        result: list[EnergySource] = []
        if sources is not None:
            for source in sources:
                flow_from = source.get("flow_from")
                flow_to = source.get("flow_to")
                if flow_from is not None and flow_to is not None:
                    energy_source = EnergySource(
                        source_type=source.get("type"),
                        flow_from=flow_from[0].get("stat_energy_from"),
                        flow_to=flow_to[0].get("stat_energy_to") if flow_to is not None else None,
                    )
                    result.append(energy_source)
                else:
                    energy_from = source.get("stat_energy_from")
                    if energy_from is not None:
                        energy_source = EnergySource(
                            source_type=source.get("type"),
                            flow_from=energy_from,
                            flow_to=None,
                        )
                        result.append(energy_source)

        return result

    async def get_solar_forecast(self) -> pd.DataFrame:
        """Get the solar forecast from Home Assistant."""
        forecast = await self.hass.send_command("energy/solar_forecast")
        df = pd.DataFrame()

        for fcst, series in forecast.items():
            df[fcst] = pd.Series(series.get("wh_hours"))
        df.index = pd.to_datetime(df.index)
        df["sum"] = df.sum(axis=1)
        freq = pd.Timedelta("60min")
        return df.resample(freq).sum().interpolate()

    async def async_read_states(self) -> None:
        """Read the states from the homeassistant instance asynchronously."""

        try:
            states = await self.hass.get_states()
            self._read_states.clear()
            for state in states:
                entity_id = state.get("entity_id")
                self._read_states[entity_id] = HomeassistantState(
                    entity_id,
                    state.get("state"),
                    state.get("attributes"),
                )
            self._template_states = None
        except Exception:
            LOGGER.exception("Exception during homeassistant update_states: ")

    async def async_write_states(self) -> None:
        """Send the changed states to hass."""
        if not self._demo_mode:
            try:
                for id, state in self._write_states.items():
                    if id.startswith("number"):
                        data = {"value": state.value}
                        await self.hass.call_service(
                            "number",
                            service="set_value",
                            service_data=data,
                            target={"entity_id": id},
                        )

                    elif id.startswith("switch"):
                        await self.hass.call_service("switch", service=f"turn_{state.value}", target={"entity_id": id})

                    elif id.startswith("sensor"):
                        headers = {
                            "Authorization": f"Bearer {self._token}",
                            "content-type": "application/json",
                        }
                        sensor_data: dict = {
                            "state": state.value,
                            "attributes": state.attributes,
                        }
                        async with self.session.post(
                            f"{self._url}/api/states/{id}",
                            headers=headers,
                            json=sensor_data,
                        ) as response:
                            if not response.ok:
                                LOGGER.error(f"State update for {id} in hass failed")
                    else:
                        LOGGER.error(f"Writing to id {id} is not yet implemented.")
            except Exception:
                LOGGER.exception("Exception during homeassistant update_states.")
            self._write_states.clear()

    def read_states(self) -> None:
        """Read the states from the homeassistant instance."""
        if self._demo_mode:
            self._read_states["sensor.solaredge_i1_ac_power"] = HomeassistantState(
                "sensor.solaredge_i1_ac_power",
                "10000",
            )
            self._read_states["sensor.solaredge_m1_ac_power"] = HomeassistantState(
                "sensor.solaredge_m1_ac_power",
                "6000",
            )
            self._read_states["sensor.keba_charge_power"] = HomeassistantState("sensor.keba_charge_power", "2500")
            self._read_states["sensor.tumbler_power"] = HomeassistantState("sensor.tumbler_power", "600")
            self._read_states["sensor.officedesk_power"] = HomeassistantState("sensor.officedesk_power", "40")
            self._read_states["sensor.rack_power"] = HomeassistantState("sensor.rack_power", "80")
        else:
            headers = {
                "Authorization": f"Bearer {self._token}",
                "content-type": "application/json",
            }
            try:
                response = requests.get(f"{self._url}/api/states", headers=headers)

                if response.ok:
                    states = response.json()
                    self._read_states.clear()
                    for state in states:
                        entity_id = state.get("entity_id")
                        self._read_states[entity_id] = HomeassistantState(
                            entity_id,
                            state.get("state"),
                            state.get("attributes"),
                        )
                    self._template_states = None

            except Exception:
                LOGGER.exception("Exception during homeassistant update_states: ")

    def write_states(self) -> None:
        """Send the changed states to hass."""
        if not self._demo_mode:
            headers = {
                "Authorization": f"Bearer {self._token}",
                "content-type": "application/json",
            }
            try:
                for id, state in self._write_states.items():
                    if id.startswith("number"):
                        data = {"entity_id": id, "value": state.value}
                        response = requests.post(
                            f"{self._url}/api/services/number/set_value",
                            headers=headers,
                            json=data,
                        )
                        if not response.ok:
                            LOGGER.error("State update in hass failed")
                    elif id.startswith("switch"):
                        data = {"entity_id": id}
                        response = requests.post(
                            f"{self._url}/api/services/switch/turn_{state.value}",
                            headers=headers,
                            json=data,
                        )
                        if not response.ok:
                            LOGGER.error("Turn switch update in hass failed")
                    elif id.startswith("sensor"):
                        sensor_data: dict = {
                            "state": state.value,
                            "attributes": state.attributes,
                        }
                        response = requests.post(
                            f"{self._url}/api/states/{id}",
                            headers=headers,
                            json=sensor_data,
                        )
                        if not response.ok:
                            LOGGER.error(f"State update for {id} in hass failed")
                    else:
                        LOGGER.error(f"Writing to id {id} is not yet implemented.")
            except Exception:
                LOGGER.exception("Exception during homeassistant update_states.")
            self._write_states.clear()

    async def get_config(self) -> dict:
        """Read the Homeassistant configuration."""
        headers = {
            "Authorization": f"Bearer {self._token}",
            "content-type": "application/json",
        }
        async with self.session.get(f"{self._url}/api/config", headers=headers) as response:
            if response.ok:
                return await response.json()
        raise HomeAssistantCommunicationError(response)

    async def get_location(self) -> Location:
        """Read the location from the Homeassistant configuration."""
        config = await self.get_config()

        return Location(
            latitude=config.get("latitude", ""),
            longitude=config.get("longitude", ""),
            elevation=config.get("elevation", ""),
            time_zone=config.get("time_zone", ""),
        )

    async def get_timezone(self) -> tzinfo:
        """Get the local timezone."""
        if self._time_zone is None:
            self._time_zone = ZoneInfo((await self.get_location()).time_zone)
        return self._time_zone


class HomeassistantDevice(DeviceWithState):
    """A generic Homeassistant device."""

    def __init__(
        self,
        device_id: uuid.UUID,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        """Create a generic Homeassistant device."""
        super().__init__(device_id, session_storage)
        self._device_type_registry = device_type_registry
        self._nominal_power: float | None = None
        self._nominal_duration: float | None = None
        self._is_constant: bool | None = None

        self._power_entity_id: str = ""
        self._power: State | None = None
        self._consumed_energy: State | None = None
        self._output_state: State | None = None
        self._output_id: str | None = None
        self._icon: str = "mdi-home"
        self._power_data = DataBuffer()

        self._state: str = OnOffState.UNKNOWN
        self._consumed_energy_value: StateValue | None = None
        self._device_type: DeviceType | None = None

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        super().configure(config)
        self._power_entity_id = get_config_param(config, "power")
        energy_config = config.get("energy")
        if energy_config is not None:
            self._consumed_energy_value = StateValue(energy_config)
        else:
            msg = "energy"
            raise DeviceConfigMissingParameterError(msg)

        energy_scale: float | None = config.get("energy_scale")
        if energy_scale is not None:
            self._consumed_energy_value.set_scale(energy_scale)
            LOGGER.warning(
                f"Homeassistant device with id {self.id} is configured with energy_scale. This is deprecated and will no longer be supported.",
            )

        self._output_id = config.get("output")
        self._icon = str(config.get("icon", "mdi-home"))

        if self._output_id is not None:
            self._supported_power_modes.add(PowerModes.PV)
            self._supported_power_modes.add(PowerModes.OPTIMIZED)

        manufacturer = config.get("manufacturer")
        model = config.get("model")

        if model is not None and manufacturer is not None:
            self._device_type = self._device_type_registry.get_device_type(manufacturer, model)
            self._nominal_power = config.get(
                "nominal_power",
                self._device_type.nominal_power if self._device_type else None,
            )
            self._nominal_duration = config.get(
                "nominal_duration",
                self._device_type.nominal_duration if self._device_type else None,
            )
            self._is_constant = config.get("constant", self._device_type.constant if self._device_type else None)
        else:
            self._nominal_power = config.get("nominal_power")
            self._nominal_duration = config.get("nominal_duration")
            self._is_constant = config.get("constant")
        state_detection: dict = config.get("state", {})
        if self._device_type is None and state_detection:
            state_on = state_detection.get("state_on", {})
            state_off = state_detection.get("state_off", {})
            self._device_type = DeviceType(
                str(config.get("icon", "mdi:lightning-bolt")),
                self._nominal_power if self._nominal_power is not None else DEFAULT_NOMINAL_POWER,
                (self._nominal_duration if self._nominal_duration is not None else DEFAULT_NOMINAL_DURATION),
                self._is_constant if self._is_constant is not None else False,
                state_on.get("threshold", 2),
                state_off.get("threshold", 0),
                state_off.get("upper", 0),
                state_off.get("lower", 0),
                state_off.get("for", 0),
                state_off.get("trailing_zeros_for", 10),
            )

    @property
    def type(self) -> str:
        """The device type."""
        return "homeassistant"

    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the own state from the states of a StatesRepository."""
        if self._output_id is not None:
            self._output_state = assign_if_available(self._output_state, state_repository.get_state(self._output_id))
        else:
            self._output_state = None
        self._power = assign_if_available(self._power, state_repository.get_state(self._power_entity_id))
        self._consumed_energy = assign_if_available(
            self._consumed_energy,
            self._consumed_energy_value.evaluate(state_repository) if self._consumed_energy_value is not None else None,
        )
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        if self.has_state:
            old_state = self.state == OnOffState.ON
            if self._device_type is not None:
                self._power_data.add_data_point(self.power)
                if self.state != OnOffState.ON and self.power > self._device_type.state_on_threshold:
                    self._state = OnOffState.ON
                elif self.state != OnOffState.OFF:
                    if self.state == OnOffState.ON and self.power <= self._device_type.state_off_threshold:
                        is_between = self._device_type.state_off_for > 0 and self._power_data.is_between(
                            self._device_type.state_off_lower,
                            self._device_type.state_off_upper,
                            self._device_type.state_off_for,
                            without_trailing_zeros=True,
                        )
                        max = self._device_type.trailing_zeros_for > 0 and self._power_data.get_max_for(
                            self._device_type.trailing_zeros_for,
                        )
                        if is_between or max <= self._device_type.state_off_threshold:
                            self._state = OnOffState.OFF
                    elif self.state == OnOffState.UNKNOWN:
                        self._state = OnOffState.OFF
            new_state = self.state == OnOffState.ON
            await super().update_session(old_state, new_state, "Power State Device")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if self._output_id is not None:
            state: bool = self._output_state.value == "on" if self._output_state is not None else False
            new_state = state
            if self.power_mode == PowerModes.PV:
                avg_300 = grid_exported_power_data.get_average_for(300)
                if avg_300 > self.nominal_power * (1 + POWER_HYSTERESIS):
                    new_state = True
                elif avg_300 < self.nominal_power * (1 - POWER_HYSTERESIS):
                    new_state = False
            elif self.power_mode == PowerModes.OPTIMIZED:
                power = optimizer.get_optimized_power(self._id)
                new_state = power > 0
            if state != new_state:
                state_repository.set_state(
                    StateId(
                        id=self._output_id,
                        channel=HOMEASSISTANT_CHANNEL,
                    ),
                    "on" if new_state else "off",
                )

    @property
    def consumed_energy(self) -> float:
        """The consumed energy of the device."""
        return self._consumed_energy.numeric_value if self._consumed_energy is not None else 0.0

    @property
    def icon(self) -> str:
        """The icon of the device."""
        return self._icon

    @property
    def power(self) -> float:
        """The current power used by the device."""
        return self._power.numeric_value if self._power else 0.0

    @property
    def available(self) -> bool:
        """Is the device available?."""
        return (
            self._consumed_energy is not None
            and self._consumed_energy.available
            and self._power is not None
            and self._power.available
        )

    @property
    def nominal_power(self) -> float:
        """The nominal power of the device."""
        return self._nominal_power if self._nominal_power is not None else DEFAULT_NOMINAL_POWER

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore a previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = HomeassistantState("", str(consumed_energy))

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is running."""
        return self._state

    @property
    def has_state(self) -> bool:
        """Has this device a state."""
        return self._device_type is not None

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if self._nominal_power is not None and self._nominal_duration is not None:
            if self.power_mode == PowerModes.OPTIMIZED:
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=self._nominal_power,
                    duration=self._nominal_duration,
                    is_continous=False,
                    is_constant=self._is_constant if self._is_constant is not None else False,
                    is_deferrable=True,
                )

            if self.state == OnOffState.ON:
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=self._nominal_power,
                    duration=self._nominal_duration - self.session_duration,
                    is_continous=False,
                    is_constant=self._is_constant if self._is_constant is not None else False,
                    is_deferrable=False,
                )

        return None



"""Data model and schema classes for a home measurement."""

from __future__ import annotations

import uuid
from collections.abc import AsyncIterator
from datetime import date
from typing import TYPE_CHECKING

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, mapped_column, relationship, selectinload

from .base import Base

if TYPE_CHECKING:
    from .device import DeviceMeasurement


def get_consumed_energy(measurement: HomeMeasurement) -> float:
    """Get the consumed energy from a measurement."""
    return measurement.solar_produced_energy + measurement.grid_imported_energy - measurement.grid_exported_energy


def get_consumed_solar_energy(measurement: HomeMeasurement) -> float:
    """Get the consumed solar energy from a measurement."""
    return measurement.solar_produced_energy - measurement.grid_exported_energy


class HomeMeasurement(Base):
    """Data model for a measurement."""

    __tablename__ = "HomeMeasurement"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    solar_produced_energy: Mapped[float]
    grid_imported_energy: Mapped[float]
    grid_exported_energy: Mapped[float]
    measurement_date: Mapped[date] = mapped_column("date")

    device_measurements: Mapped[list[DeviceMeasurement]] = relationship(
        "DeviceMeasurement",
        back_populates="home_measurement",
        order_by="DeviceMeasurement.id",
        cascade="save-update, merge, refresh-expire, expunge, delete, delete-orphan",
    )

    @property
    def consumed_energy(self) -> float:
        """Get the consumed energy from a measurement."""
        return self.solar_produced_energy + self.grid_imported_energy - self.grid_exported_energy

    @property
    def solar_consumed_energy(self) -> float:
        """Get the consumed solar energy from a measurement."""
        return self.solar_produced_energy - self.grid_exported_energy

    @classmethod
    async def read_all(cls, session: AsyncSession, include_device_measurements: bool) -> AsyncIterator[HomeMeasurement]:
        """Read all home measurements."""
        stmt = select(cls)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        stream = await session.stream_scalars(stmt.order_by(cls.measurement_date.desc()))
        async for row in stream:
            yield row

    @classmethod
    async def read_by_id(
        cls,
        session: AsyncSession,
        home_measurement_id: int,
        include_device_measurements: bool = False,
    ) -> HomeMeasurement | None:
        """Read a home measurements by id."""
        stmt = select(cls).where(cls.id == home_measurement_id)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        return await session.scalar(stmt.order_by(cls.id))

    @classmethod
    async def read_first(
        cls,
        session: AsyncSession,
        include_device_measurements: bool = False,
    ) -> HomeMeasurement | None:
        """Read last home measurement by date."""
        stmt = select(cls)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        return await session.scalar(stmt.order_by(cls.measurement_date).limit(1))

    @classmethod
    async def read_last(
        cls,
        session: AsyncSession,
        include_device_measurements: bool = False,
    ) -> HomeMeasurement | None:
        """Read last home measurement."""
        stmt = select(cls)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        return await session.scalar(stmt.order_by(cls.measurement_date.desc()).limit(1))

    @classmethod
    async def read_by_date(
        cls,
        session: AsyncSession,
        measurement_date: date,
        include_device_measurements: bool = False,
    ) -> HomeMeasurement | None:
        """Read last home measurement by date."""
        stmt = select(cls).where(cls.measurement_date == measurement_date)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        return await session.scalar(stmt.order_by(cls.measurement_date.desc()).limit(1))

    @classmethod
    async def read_between_dates(
        cls,
        session: AsyncSession,
        from_date: date,
        to_date: date,
        include_device_measurements: bool = False,
    ) -> AsyncIterator[HomeMeasurement]:
        """Read last home measurement by date."""
        stmt = select(cls).where(cls.measurement_date >= from_date).where(cls.measurement_date <= to_date)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        stream = await session.stream_scalars(stmt.order_by(cls.measurement_date))
        async for row in stream:
            yield row

    @classmethod
    async def read_before_date(
        cls,
        session: AsyncSession,
        measurement_date: date,
        include_device_measurements: bool = False,
    ) -> HomeMeasurement | None:
        """Read last home measurement by date."""
        stmt = select(cls).where(cls.measurement_date < measurement_date)
        if include_device_measurements:
            stmt = stmt.options(selectinload(cls.device_measurements))
        return await session.scalar(stmt.order_by(cls.measurement_date.desc()).limit(1))

    @classmethod
    async def create(
        cls,
        session: AsyncSession,
        name: str,
        solar_produced_energy: float,
        grid_imported_energy: float,
        grid_exported_energy: float,
        measurement_date: date,
        device_measurements: list[DeviceMeasurement],
    ) -> HomeMeasurement:
        """Create a home measurement."""
        home_measurement = HomeMeasurement(
            name=name,
            solar_produced_energy=solar_produced_energy,
            grid_imported_energy=grid_imported_energy,
            grid_exported_energy=grid_exported_energy,
            measurement_date=measurement_date,
            device_measurements=device_measurements,
        )
        session.add(home_measurement)
        await session.flush()
        # To fetch device measurements
        new = await cls.read_by_id(session, home_measurement.id, include_device_measurements=True)
        if not new:
            raise RuntimeError
        return new

    async def update(
        self,
        session: AsyncSession,
        name: str,
        solar_produced_energy: float,
        grid_imported_energy: float,
        grid_exported_energy: float,
        measurement_date: date,
    ) -> None:
        """Update a home measurement."""
        self.name = name
        self.solar_produced_energy = solar_produced_energy
        self.grid_imported_energy = grid_imported_energy
        self.grid_exported_energy = grid_exported_energy
        self.measurement_date = measurement_date
        await session.flush()

    @classmethod
    async def delete(cls, session: AsyncSession, home_measurement: HomeMeasurement) -> None:
        """Delete a home measurement."""
        await session.delete(home_measurement)
        await session.flush()

    def get_device_measurement(self, device_id: uuid.UUID) -> DeviceMeasurement | None:
        """Find the device measurement of the device with the given id."""
        for device_measurement in self.device_measurements:
            if device_measurement.device_id == device_id:
                return device_measurement
        return None



"""Config helper classes and functions."""

from energy_assistant.devices import Location
from energy_assistant.storage.config import ConfigStorage, DeviceConfigMissingParameterError


def get_config_param(config: dict, param: str) -> str:
    """Get a config parameter as string or raise an exception if the parameter is not available."""
    result = config.get(param)
    if result is None:
        raise DeviceConfigMissingParameterError(param)
    return str(result)


def get_config_param_from_list(config: list, param: str) -> str | None:
    """Read config param from a list."""
    for item in config:
        value = item.get(param)
        if value is not None:
            return value
    return None


def get_float_param_from_list(config: list, param: str) -> float | None:
    """Read a float config param from a list."""
    for item in config:
        value = item.get(param)
        if value is not None:
            return float(value)
    return None


class EnergyAssistantConfig:
    """The energy assistant config."""

    def __init__(self, energy_assistant_config: ConfigStorage, hass_config: dict) -> None:
        """Create a EnergyAssistantConfig instance."""
        self._energy_assistant_config = energy_assistant_config
        self._hass_config = hass_config

    @property
    def as_dict(self) -> dict:
        """Get the complete config."""
        result = {
            "energy_assistant": self._energy_assistant_config.as_dict(),
            "home_assistant": self._hass_config,
        }
        result["emhass"] = self._energy_assistant_config.emhass.as_dict()
        del result["energy_assistant"]["emhass"]
        return result

    @property
    def energy_assistant_config(self) -> ConfigStorage:
        """Get the energy assistant config."""
        return self._energy_assistant_config

    @property
    def home_assistant_config(self) -> dict:
        """Get the home assistant config."""
        return self._hass_config

    @property
    def emhass_config(self) -> dict:
        """Get the emhass config."""
        return self._energy_assistant_config.emhass.as_dict()

    @property
    def location(self) -> Location:
        """Read the location from the Homeassistant configuration."""
        config = self.home_assistant_config

        return Location(
            latitude=config.get("latitude", ""),
            longitude=config.get("longitude", ""),
            elevation=config.get("elevation", ""),
            time_zone=config.get("time_zone", ""),
        )



"""The application."""

import uuid
from abc import ABC, abstractmethod

__version__ = "0.1.17"


class Optimizer(ABC):
    """Base class for optimizers."""

    @abstractmethod
    def get_optimized_power(self, device_id: uuid.UUID) -> float:
        """Get the optimized power budget for a give device."""



"""Main module for the energy assistant application."""

import asyncio
import json
import logging
import os
import sys
import threading
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager, suppress
from datetime import UTC, datetime, tzinfo
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Final

import alembic.config
import pandas as pd
from anyio import open_file
from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore
from colorlog import ColoredFormatter
from energy_assistant_frontend import where as locate_frontend
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker

from energy_assistant.api.main import router as api_router
from energy_assistant.devices import StatesMultipleRepositories, StatesRepository
from energy_assistant.devices.config import EnergyAssistantConfig
from energy_assistant.devices.evcc import EvccDevice
from energy_assistant.devices.home import Home
from energy_assistant.devices.homeassistant import Homeassistant
from energy_assistant.devices.registry import DeviceTypeRegistry
from energy_assistant.emhass_optimizer import EmhassOptimizer
from energy_assistant.importer.homeassistant import import_data
from energy_assistant.mqtt import MqttConnection
from energy_assistant.settings import settings
from energy_assistant.storage.config import ConfigStorage
from energy_assistant.storage.storage import Database, get_async_session, session_storage
from energy_assistant.websocket import get_home_message, ws_manager, ws_router

from .constants import ROOT_LOGGER_NAME

FORMAT_DATE: Final = "%Y-%m-%d"
FORMAT_TIME: Final = "%H:%M:%S"
FORMAT_DATETIME: Final = f"{FORMAT_DATE} {FORMAT_TIME}"
MAX_LOG_FILESIZE = 1000000 * 10  # 10 MB


class EnergyAssistant:
    """Energy Assistant Application."""

    home: Home
    hass: Homeassistant | None
    optimizer: EmhassOptimizer
    mqtt: MqttConnection | None
    db: Database
    config: EnergyAssistantConfig
    should_stop = False


async def import_data_task(
    home: Home,
    hass: Homeassistant,
    async_session: async_sessionmaker,
    freq: pd.Timedelta,
    days_to_retrieve: int,
) -> None:
    """Import data from home assistant."""
    async with async_session() as session:
        await import_data(home, hass, session, freq, days_to_retrieve)


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator:
    """Manage the startup and showdown."""
    ea = await init_app()
    app.energy_assistant = ea  # type: ignore
    bt = asyncio.create_task(background_task(ea))
    optimizer_task = asyncio.create_task(optimize(ea.optimizer))
    importer_task = None
    if ea.hass is not None:
        importer_task = asyncio.create_task(
            import_data_task(ea.home, ea.hass, await get_async_session(), pd.Timedelta(24, "h"), 8),
        )
    scheduler = AsyncIOScheduler()
    scheduler.add_job(async_daily_optimize, trigger="cron", args=[ea], hour="3", minute="0")  # time is UTC
    scheduler.start()
    yield
    ea.should_stop = True
    await optimizer_task
    if importer_task is not None:
        await importer_task
    await bt
    scheduler.shutdown()
    if ea.hass:
        await ea.hass.disconnect()
    print("Shutdown app")


app = FastAPI(title="energy-assistant", lifespan=lifespan)


origins = [
    "http://localhost",
    "http://localhost:5000",
    "http://localhost:5173",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")
app.include_router(ws_router)


async def async_handle_state_update(
    ea: EnergyAssistant,
    state_repository: StatesRepository,
    session: AsyncSession,
    time_zone: tzinfo,
) -> None:
    """Read the values from home assistant and process the update."""
    try:
        await state_repository.async_read_states()

        await ea.home.update_state(state_repository)
        if ea.optimizer is not None:
            await ea.home.update_power_consumption(state_repository, ea.optimizer)
            ea.optimizer.update_repository_states(ea.home, state_repository)
        else:
            logging.error("The variable optimizer is None in async_handle_state_update")
        await state_repository.async_write_states()
        if ea.db:
            if ea.home:
                await asyncio.gather(
                    ws_manager.broadcast(get_home_message(ea.home)),
                    ea.db.store_home_state(ea.home, session, time_zone),
                )
            else:
                logging.error("The variable home is None in async_handle_state_update")
        else:
            logging.error("The variable db is None in async_handle_state_update")
        if ea.optimizer is not None and ea.home is not None:
            await ea.optimizer.async_update_devices(ea.home)
    except Exception:
        logging.exception("error during sending refresh")


async def background_task(ea: EnergyAssistant) -> None:
    """Periodically read the values from home assistant and process the update."""
    time_zone = await ea.hass.get_timezone() if ea.hass is not None else UTC
    last_update = datetime.now(tz=time_zone).date()
    async_session = await get_async_session()
    repositories: list[StatesRepository] = []
    if ea.hass is not None:
        repositories.append(ea.hass)
    if ea.mqtt is not None:
        repositories.append(ea.mqtt)
    state_repository = StatesMultipleRepositories(repositories)
    while not ea.should_stop:
        # delta_t = datetime.now().timestamp()
        # print("Start refresh from home assistant")
        today = datetime.now(tz=time_zone).date()
        try:
            if today != last_update:
                ea.home.store_energy_snapshot()
            last_update = today

            async with async_session() as session:
                await async_handle_state_update(ea, state_repository, session, time_zone)
        except Exception:
            logging.exception("error in the background task")
        # print(f"refresh from home assistant completed in {datetime.now().timestamp() - delta_t} s")
        await asyncio.sleep(30)


def create_mqtt_connection(config: ConfigStorage) -> MqttConnection | None:
    """Create an mqtt connection based on the config."""

    mqtt_config = config.mqtt
    if mqtt_config is not None:
        mqtt_host = mqtt_config.get("host")
        mqtt_username = mqtt_config.get("username")
        mqtt_password = mqtt_config.get("password")
        mqtt_topic = mqtt_config.get("topic")
        mqtt_connection = MqttConnection(mqtt_host, mqtt_username, mqtt_password, mqtt_topic)
        mqtt_connection.connect()
        return mqtt_connection
    return None


def subscribe_mqtt_topics(mqtt_connection: MqttConnection, home: Home) -> None:
    """Subscribe the mqtt based devices on the mqtt connection."""

    for device in home.devices:
        if isinstance(device, EvccDevice):
            mqtt_connection.add_subscription_topic(device.evcc_mqtt_subscription_topic)


async def open_hass_connection(config: ConfigStorage) -> Homeassistant | None:
    """Create a connection to home assistant."""
    token: str | None = None
    url: str | None = None
    try:
        token = os.getenv("SUPERVISOR_TOKEN")
        if token is not None:
            logging.info(f"suvervisor token detected. len={len(token)}")
            url = "http://supervisor/core"

            hass = Homeassistant(url, token, False)
            await hass.connect()
            return hass
    except Exception:
        logging.exception("Error while trying to connect to the homeassistant supervisor api")
        url = None
        token = None

    logging.info("Try to connect to home assistant based on the config file entries...")
    hass_config = config.homeassistant
    if hass_config is not None:
        demo_mode = hass_config.get("demo_mode")
        url = hass_config.get("url")
        token = hass_config.get("token")
        if url is not None and token is not None:
            hass = Homeassistant(url, token, demo_mode)
            await hass.connect()
            return hass
    return None


def setup_logger(log_filename: str, level: str = "DEBUG") -> logging.Logger:
    """Initialize logger."""
    # define log formatter
    log_fmt = "%(asctime)s.%(msecs)03d %(levelname)s (%(threadName)s) [%(name)s] %(message)s"

    # base logging config for the root logger
    logging.basicConfig(level=logging.INFO)

    colorfmt = f"%(log_color)s{log_fmt}%(reset)s"
    logging.getLogger().handlers[0].setFormatter(
        ColoredFormatter(
            colorfmt,
            datefmt=FORMAT_DATETIME,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red",
            },
        ),
    )

    # Capture warnings.warn(...) and friends messages in logs.
    # The standard destination for them is stderr, which may end up unnoticed.
    # This way they're where other messages are, and can be filtered as usual.
    logging.captureWarnings(True)

    # setup file handler
    # log_filename = os.path.join(data_path, "energy_assistant.log")
    file_handler = RotatingFileHandler(log_filename, maxBytes=MAX_LOG_FILESIZE, backupCount=1)
    # rotate log at each start
    with suppress(OSError):
        file_handler.doRollover()
    file_handler.setFormatter(logging.Formatter(log_fmt, datefmt=FORMAT_DATETIME))
    # file_handler.setLevel(logging.INFO)

    logger = logging.getLogger()
    logger.addHandler(file_handler)

    # apply the configured global log level to the (root) music assistant logger
    logging.getLogger(ROOT_LOGGER_NAME).setLevel(level)

    # silence some noisy loggers
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("aiosqlite").setLevel(logging.WARNING)
    logging.getLogger("databases").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    sys.excepthook = lambda *args: logging.getLogger(None).exception(
        "Uncaught exception",
        exc_info=args,  # type: ignore[arg-type]
    )
    threading.excepthook = lambda args: logging.getLogger(None).exception(
        "Uncaught thread exception",
        exc_info=(  # type: ignore[arg-type]
            args.exc_type,
            args.exc_value,
            args.exc_traceback,
        ),
    )

    return logger


async def init_app() -> EnergyAssistant:
    """Initialize the application."""
    result = EnergyAssistant()

    hass_options_file = Path("/data/options.json")
    if hass_options_file.is_file():
        async with await open_file(hass_options_file) as _file:
            hass_options = json.loads(await _file.read())
    else:
        hass_options = {}

    log_level = hass_options.get("log_level", "info").upper()

    logger = setup_logger(settings.LOG_FILE, log_level)

    config_file = hass_options.get("config_file", settings.CONFIG_FILE)

    logger.info("Hello from Energy Assistant")

    for option in hass_options:
        logging.info(f"{option}={hass_options[option]}")

    async_session = await get_async_session()
    db = Database()
    result.db = db

    device_registry_path = Path(__file__).parent / "config/deviceregistry"

    device_type_registry = DeviceTypeRegistry()
    device_type_registry.load(device_registry_path)

    logger.info(f"Loading config file {config_file}")
    try:
        config = ConfigStorage(Path(settings.DATA_FOLDER))
        await config.initialize(config_file)
        hass = await open_hass_connection(config)
        result.hass = hass
        result.config = EnergyAssistantConfig(config, await hass.get_config() if hass is not None else {})
        if hass is not None:
            hass.read_states()
            optimizer = EmhassOptimizer(settings.DATA_FOLDER, config, hass, await hass.get_location())
            result.optimizer = optimizer
            app.optimizer = optimizer  # type: ignore

        mqtt_connection: MqttConnection | None = create_mqtt_connection(config)
        result.mqtt = mqtt_connection

        home_config = config.home
        if home_config is not None and home_config.get("name") is not None:
            home = Home(config, session_storage, device_type_registry)
            result.home = home
            app.home = home  # type: ignore
            if mqtt_connection is not None:
                subscribe_mqtt_topics(mqtt_connection, home)

            async with async_session() as session:
                await db.update_devices(home, session, session_storage, device_type_registry)

                await db.restore_home_state(home, session)
        else:
            logger.error(f"home not found in config file: {config}")
        logger.info("Initialization completed")
    except Exception:
        logger.exception("Initialization of the app failed")
    return result


async def optimize(optimizer: EmhassOptimizer) -> None:
    """Optimize the forecast."""
    try:
        optimizer.forecast_model_fit(True)
    except Exception:
        logging.exception(
            "Optimization of the power consumption forecast model failed, probably due to missing history data in Home Assistant.",
        )


async def async_daily_optimize(ea: EnergyAssistant) -> None:
    """Optimize once a day."""
    try:
        optimizer = ea.optimizer
        if optimizer is not None:
            logging.info("Start optimizer run")
            await optimizer.async_dayahead_forecast_optim()
    except Exception:
        logging.exception("Daily optimization run failed")


@app.get("/check", include_in_schema=False)
async def health() -> JSONResponse:
    """Test the web server with a ping."""
    return JSONResponse({"message": "It worked!!"})


# This needs to be the last mount
frontend_dir = locate_frontend()
app.mount("/", StaticFiles(directory=frontend_dir, html=True), name="frontend")


def main() -> None:
    """Start energy assistant."""
    import uvicorn

    try:
        alembic_config = Path(__file__).parent / "alembic.ini"

        alembic_args = [
            "-c",
            str(alembic_config),
            "--raiseerr",
            "upgrade",
            "head",
        ]
        alembic.config.main(argv=alembic_args)  # type: ignore
    except Exception:
        print("Alembic Migration failed")
        logging.exception("Alembic Migration failed")

    uvicorn.run(app, host="0.0.0.0", port=5000)


if __name__ == "__main__":
    main()



"""Device base class for all devices."""

import logging
import uuid
from abc import ABC, abstractmethod
from datetime import UTC, datetime

from energy_assistant import Optimizer
from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.devices.analysis import DataBuffer
from energy_assistant.devices.utility_meter import UtilityMeter

from . import (
    EnergyIntegrator,
    EnergySnapshot,
    LoadInfo,
    PowerModes,
    Session,
    SessionStorage,
    StatesRepository,
)
from .config import get_config_param

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class Device(ABC):
    """A device which tracks energy consumption."""

    def __init__(self, device_id: uuid.UUID) -> None:
        """Create a device."""
        self._id = device_id
        self._consumed_solar_energy = EnergyIntegrator()
        self._energy_snapshot: EnergySnapshot | None = None
        self._supported_power_modes: set[PowerModes] = {PowerModes.DEVICE_CONTROLLED}
        self._power_mode: PowerModes = PowerModes.DEVICE_CONTROLLED
        self._utility_meters: list[UtilityMeter] = []
        self._config: dict = {}
        self._name: str = ""

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        self._name = get_config_param(config, "name")
        self._config = config.copy()

    @property
    def name(self) -> str:
        """The name of the device."""
        return self._name

    @property
    @abstractmethod
    def type(self) -> str:
        """The device type."""

    @property
    def supported_power_modes(self) -> list[PowerModes]:
        """Returns the supported power modes for the device."""
        return list(self._supported_power_modes)

    @property
    def power_mode(self) -> PowerModes:
        """The power mode of the device."""
        return self._power_mode

    @property
    def power_controllable(self) -> bool:
        """The power mode of the device."""
        return not (
            len(self._supported_power_modes) == 1 and PowerModes.DEVICE_CONTROLLED in self._supported_power_modes
        )

    @property
    def config(self) -> dict:
        """Return the config dictionary."""
        return self._config

    def set_power_mode(self, power_mode: PowerModes) -> None:
        """Set the power mode of the device."""
        self._power_mode = power_mode

    @property
    def id(self) -> uuid.UUID:
        """The id of the device."""
        return self._id

    @property
    def consumed_solar_energy(self) -> float:
        """Consumed solar energy in kWh."""
        return (
            self._consumed_solar_energy.consumed_solar_energy
            if self._consumed_solar_energy.consumed_solar_energy is not None
            else 0.0
        )

    @property
    @abstractmethod
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""

    @property
    @abstractmethod
    def power(self) -> float:
        """Current consumed power."""

    @property
    @abstractmethod
    def available(self) -> float:
        """Is the device available?."""

    @property
    @abstractmethod
    def icon(self) -> str:
        """Icon for the device."""

    @abstractmethod
    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""

    @abstractmethod
    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the state of the device."""

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore a previously stored state of the device."""
        self._consumed_solar_energy.restore_state(consumed_solar_energy, consumed_energy)
        self.set_snapshot(consumed_solar_energy, consumed_energy)

    def set_snapshot(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Set the snapshot values."""
        self._energy_snapshot = EnergySnapshot(consumed_solar_energy, consumed_energy)

    @property
    def energy_snapshot(self) -> EnergySnapshot | None:
        """The last energy snapshot of the device."""
        return self._energy_snapshot

    def store_energy_snapshot(self) -> None:
        """Store the current values in the snapshot."""
        self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        return {}

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        return None

    def add_utility_meter(self, id: str) -> UtilityMeter:
        """Add a new utility meter to the device."""
        utility_meter = UtilityMeter(id)
        self._utility_meters.append(utility_meter)
        return utility_meter


class DeviceWithState(Device):
    """Device with a state."""

    def __init__(self, device_id: uuid.UUID, session_storage: SessionStorage) -> None:
        """Create a DeviceWithState instance."""
        super().__init__(device_id)
        self.session_storage: SessionStorage = session_storage
        self.current_session: Session | None = None
        self._store_sessions: bool = False

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        super().configure(config)
        self._store_sessions = config.get("store_sessions", False)

    @property
    @abstractmethod
    def state(self) -> str:
        """The state of the device."""

    @property
    def has_state(self) -> bool:
        """Has this device a state."""
        return True

    @property
    def session_duration(self) -> float:
        """Return the duration of the current session."""
        if self.current_session is not None:
            result = datetime.now(UTC) - self.current_session.start
            return result.total_seconds()
        return 0.0

    async def start_session(self, text: str) -> None:
        """Start a session."""
        self.current_session = await self.session_storage.start_session(
            self.id,
            text,
            self.consumed_solar_energy,
            self.consumed_energy,
        )

    async def update_session(self, old_state: bool, new_state: bool, text: str) -> None:
        """Update the session log."""
        if self._store_sessions:
            if new_state:
                if old_state:
                    if self.current_session is not None:
                        await self.session_storage.update_session(
                            self.current_session.id,
                            self.consumed_solar_energy,
                            self.consumed_energy,
                        )
                else:
                    LOGGER.info("Start Session")
                    await self.start_session(text)
            else:
                if old_state:
                    LOGGER.info("End Session")
                if self.current_session is not None:
                    await self.session_storage.update_session_energy(
                        self.current_session.id,
                        self.consumed_solar_energy,
                        self.consumed_energy,
                    )

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        result: dict[str, str] = {}
        if self.has_state:
            result["state"] = self.state
        if self.state == "on" and self.current_session is not None:
            result["session_time"] = str((datetime.now(UTC) - self.current_session.start).total_seconds())
            result["session_energy"] = str(self.consumed_energy - self.current_session.start_consumed_energy)
            result["session_solar_energy"] = str(
                self.consumed_solar_energy - self.current_session.start_solar_consumed_energy,
            )
        return result



"""remove consumed energy columns

Revision ID: fa8f65c2f997
Revises: 2558c9508fcc
Create Date: 2024-06-11 15:34:29.013567

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "fa8f65c2f997"
down_revision = "2558c9508fcc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("HomeMeasurement") as batch_op:
        batch_op.drop_column("consumed_energy")
        batch_op.drop_column("solar_consumed_energy")


def downgrade() -> None:
    op.add_column("HomeMeasurement", sa.Column("consumed_energy", sa.FLOAT(), nullable=False))
    op.add_column("HomeMeasurement", sa.Column("solar_consumed_energy", sa.FLOAT(), nullable=False))



"""All constants for Energy Assistant."""

from typing import Final

ROOT_LOGGER_NAME: Final[str] = "energy_assistant"

DEFAULT_NOMINAL_POWER = 100
DEFAULT_NOMINAL_DURATION = 3600

POWER_HYSTERESIS = 0.1



"""Views for home measurement API."""

from datetime import date
from typing import Annotated

from fastapi import APIRouter, Depends, Request

from .schema import (
    HomeMeasurementDailyResponse,
    HomeMeasurementDailySchema,
    HomeMeasurementPeriodSchema,
    ReadHomeMeasurementDifferenceResponse,
)
from .use_cases import ReadHomeMeasurementDaily, ReadHomeMeasurementDifference

router = APIRouter(prefix="/history")


@router.get(
    "/difference",
    response_model=ReadHomeMeasurementDifferenceResponse,
)
async def read_difference(
    request: Request,
    from_date: date,
    to_date: date,
    use_case: Annotated[ReadHomeMeasurementDifference, Depends(ReadHomeMeasurementDifference)],
) -> HomeMeasurementPeriodSchema:
    """Get the difference of the measurements between to dates."""
    return await use_case.execute(from_date, to_date)


@router.get(
    "/daily",
    response_model=HomeMeasurementDailyResponse,
)
async def read_daily(
    request: Request,
    from_date: date,
    to_date: date,
    use_case: Annotated[ReadHomeMeasurementDaily, Depends(ReadHomeMeasurementDaily)],
) -> HomeMeasurementDailySchema:
    """Get the difference of the measurements between to dates."""
    return await use_case.execute(from_date, to_date)



"""Helper classes for data analysis."""

import pathlib
from collections import deque
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta, tzinfo
from statistics import mean

import pandas as pd

MAX_DATA_LEN = 3000  # in case of 30 seconds interval, this is about one day.


class OnOffDataBuffer:
    """Data buffer for OnOff States."""

    def __init__(self) -> None:
        """Create on/off data buffer instance."""
        self.data = pd.Series()

    def add_data_point(self, value: bool, time_stamp: datetime | None = None) -> None:
        """Add a new data point for tracking."""
        if time_stamp is None:
            time_stamp = datetime.now(UTC)
        self.data[time_stamp] = value
        if len(self.data) > MAX_DATA_LEN:
            self.data = self.data.tail(MAX_DATA_LEN)


@dataclass
class DataPoint:
    """Data point for the DataBuffer."""

    value: float
    time_stamp: datetime


class DataBuffer:
    """Data buffer for analysis."""

    def __init__(self) -> None:
        """Create a DataBuffer instance."""
        self.data: deque = deque([], MAX_DATA_LEN)

    def add_data_point(self, value: float, time_stamp: datetime | None = None) -> None:
        """Add a new data point for tracking."""
        if time_stamp is None:
            time_stamp = datetime.now(UTC)
        self.data.append(DataPoint(value, time_stamp))

    def get_data_for(
        self,
        timespan: float,
        now: datetime | None = None,
        without_trailing_zeros: bool = False,
    ) -> list[float]:
        """Extract data for the last timespan seconds."""
        if now is None:
            now = datetime.now(UTC)
        threshold = now - timedelta(seconds=timespan)
        result = [data_point.value for data_point in self.data if data_point.time_stamp >= threshold]
        if len(result) == 0:
            result = [self.data[-1].value]
        if without_trailing_zeros:
            while result[-1] == 0.0:
                result.pop()
        return result

    def get_average_for(self, timespan: float, now: datetime | None = None) -> float:
        """Calculate the average over the last timespan seconds."""
        if now is None:
            now = datetime.now(UTC)
        return mean(self.get_data_for(timespan, now))

    def average(self) -> float:
        """Average of the data buffer."""
        if len(self.data) > 0:
            return mean([d.value for d in self.data])
        return 0.0

    def get_min_for(self, timespan: float, now: datetime | None = None) -> float:
        """Calculate the min over the last timespan seconds."""
        if now is None:
            now = datetime.now(UTC)
        return min(self.get_data_for(timespan, now))

    def get_max_for(self, timespan: float, now: datetime | None = None) -> float:
        """Calculate the max over the last timespan seconds."""
        if now is None:
            now = datetime.now(UTC)
        return max(self.get_data_for(timespan, now))

    def is_between(
        self,
        lower: float,
        upper: float,
        timespan: float,
        now: datetime | None = None,
        without_trailing_zeros: bool = False,
    ) -> bool:
        """Check if the value in the timespan is always between lower and upper."""
        if now is None:
            now = datetime.now(UTC)
        data = self.get_data_for(timespan, now, without_trailing_zeros)
        if len(data) > 0:
            if min(data) < lower:
                return False
            return max(data) <= upper
        return False

    def get_data_frame(
        self,
        freq: pd.Timedelta,
        time_zone: tzinfo,
        value_name: str,
        folder: pathlib.Path,
    ) -> pd.DataFrame:
        """Get a pandas data frame from from the available data."""
        data = [(pd.to_datetime(d.time_stamp, utc=True), d.value) for d in self.data]
        result = pd.DataFrame.from_records(data, index="timestamp", columns=["timestamp", value_name])
        result.to_csv(folder / f"{value_name}.csv")
        if not result.empty:
            result.index = result.index.tz_convert(time_zone)  # type: ignore
            return result.resample(freq).mean()
        return result


def create_timeseries_from_const(
    value: float,
    duration: pd.Timedelta,
    freq: pd.Timedelta,
    start: pd.Timestamp | None = None,
) -> pd.Series:
    """Create a time series with constant values."""
    if start is None:
        start = pd.Timestamp.utcnow()
    data = [value, value]
    index = [start, start + duration]
    result = pd.Series(data=data, index=index)
    return result.resample(freq).bfill()



"""Helper classes for web socket."""

import json
import logging

from fastapi import APIRouter, WebSocket, WebSocketDisconnect

from energy_assistant.api.device import OTHER_DEVICE
from energy_assistant.devices.device import Device
from energy_assistant.devices.home import Home

from .constants import ROOT_LOGGER_NAME

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)

ws_router = APIRouter()


class WebSocketConnectionManager:
    """Web Socket connection manager."""

    def __init__(self) -> None:
        """Initialize the web socket connection manager instance."""
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket) -> None:
        """Connect handler."""
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket) -> None:
        """Disconnect handler."""
        self.active_connections.remove(websocket)

    async def send_personal_message(self, message: str, websocket: WebSocket) -> None:
        """Send a message to one web socket."""
        await websocket.send_text(message)

    async def broadcast(self, message: str) -> None:
        """Broadcast a message."""
        for connection in self.active_connections:
            await connection.send_text(message)


ws_manager = WebSocketConnectionManager()


def get_self_sufficiency(consumed_solar_energy: float, consumed_energy: float) -> float:
    """Calculate the self sufficiency value."""
    return min(round(consumed_solar_energy / consumed_energy * 100) if consumed_energy > 0 else 0.0, 100)


def get_self_consumption(produced_solar_energy: float, consumed_solar_energy: float) -> float:
    """Calculate the self sufficiency value."""
    return min(
        round(consumed_solar_energy / produced_solar_energy * 100) if produced_solar_energy > 0 else 0.0,
        100,
    )


def get_device_message(device: Device) -> dict:
    """Generate the update data message for a device."""
    if device.energy_snapshot is not None:
        consumed_energy_today = device.consumed_energy - device.energy_snapshot.consumed_energy
        consumed_solar_energy_today = device.consumed_solar_energy - device.energy_snapshot.consumed_solar_energy
    else:
        consumed_energy_today = 0
        consumed_solar_energy_today = 0
    result = {
        "device_id": str(device.id),
        "type": device.__class__.__name__,
        "power": device.power,
        "available": device.available,
        "today": {
            "consumed_solar_energy": consumed_solar_energy_today,
            "consumed_energy": consumed_energy_today,
            "self_sufficiency": get_self_sufficiency(consumed_solar_energy_today, consumed_energy_today),
        },
    }
    result["attributes"] = device.attributes
    return result


def get_home_message(home: Home) -> str:
    """Generate the update data message for a home."""
    devices_messages = []
    if home.energy_snapshop is not None:
        consumed_energy_today = home.consumed_energy - home.energy_snapshop.consumed_energy
        consumed_solar_energy_today = home.consumed_solar_energy - home.energy_snapshop.consumed_solar_energy
        produced_solar_energy_today = home.produced_solar_energy - home.energy_snapshop.produced_solar_energy
    else:
        consumed_energy_today = 0
        consumed_solar_energy_today = 0
        produced_solar_energy_today = 0

    other_power = home.home_consumption_power
    other_consumed_energy = consumed_energy_today
    other_consumed_solar_energy = consumed_solar_energy_today
    for device in home.devices:
        device_message = get_device_message(device)
        devices_messages.append(device_message)
        other_power = max(other_power - device.power, 0.0)
        other_consumed_energy = max(other_consumed_energy - device_message["today"]["consumed_energy"], 0.0)
        other_consumed_solar_energy = max(
            other_consumed_solar_energy - device_message["today"]["consumed_solar_energy"],
            0.0,
        )

    other_device = {
        "device_id": str(OTHER_DEVICE),
        "type": "other_device",
        "power": other_power,
        "available": True,
        "today": {
            "consumed_solar_energy": other_consumed_solar_energy,
            "consumed_energy": other_consumed_energy,
            "self_sufficiency": get_self_sufficiency(other_consumed_solar_energy, other_consumed_energy),
        },
    }
    devices_messages.append(other_device)

    home_message = {
        "name": home.name,
        "power": {
            "solar_production": home.solar_production_power,
            "grid_supply": home.grid_imported_power,
            "solar_self_consumption": home.solar_self_consumption_power,
            "home_consumption": home.home_consumption_power,
            "self_sufficiency": round(home.self_sufficiency * 100),
            "self_consumption": round(home.self_consumption * 100),
        },
        "today": {
            "consumed_solar_energy": consumed_solar_energy_today,
            "consumed_energy": consumed_energy_today,
            "produced_solar_energy": produced_solar_energy_today,
            "self_sufficiency": get_self_sufficiency(consumed_solar_energy_today, consumed_energy_today),
            "self_consumption": get_self_consumption(produced_solar_energy_today, consumed_solar_energy_today),
        },
        "devices": devices_messages,
    }
    return json.dumps(home_message)


@ws_router.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket) -> None:
    """Web Socket end point for broad casts."""
    await ws_manager.connect(websocket)
    try:
        if hasattr(websocket.app, "energy_assistant"):
            ea = websocket.app.energy_assistant  # type: ignore
            await ws_manager.broadcast(get_home_message(ea.home))

        while True:
            data = await websocket.receive_text()
            LOGGER.error(f"received unexpected data from frontend: {data}")
    except WebSocketDisconnect:
        ws_manager.disconnect(websocket)



"""Use cases for devices."""

import logging
import uuid

from fastapi import HTTPException

from energy_assistant.api.config.schema import ConfigModel
from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.db import AsyncSession
from energy_assistant.devices.config import EnergyAssistantConfig
from energy_assistant.devices.home import Home
from energy_assistant.models.device import Device

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class ReadConfiguration:
    """Read the configuration use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read configuration use case."""
        self.async_session = session

    async def execute(self, config: EnergyAssistantConfig) -> ConfigModel:
        """Execute the read configuration use case."""
        return ConfigModel.model_validate({"config": config.as_dict})


class ReadDeviceConfiguration:
    """Read the configuration use case."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a read configuration use case."""
        self.async_session = session

    async def execute(self, config: EnergyAssistantConfig, device_id: uuid.UUID) -> ConfigModel:
        """Execute the read configuration use case."""
        return ConfigModel.model_validate(
            {"config": config.energy_assistant_config.devices.get_device_config(device_id)}
        )


class WriteDeviceConfiguration:
    """Write the configuration values on a device."""

    def __init__(self, session: AsyncSession) -> None:
        """Create a write configuration use case."""
        self.async_session = session

    async def execute(self, config: EnergyAssistantConfig, data: dict, device_id: uuid.UUID, home: Home) -> ConfigModel:
        """Execute the write configuration use case."""
        async with self.async_session.begin() as session:
            device = home.get_device(device_id)
            persisted_device = await Device.read_by_id(session, device_id)
            if device is None or persisted_device is None:
                raise HTTPException(status_code=404)

            for key, value in data.items():
                config.energy_assistant_config.devices.set(device_id, key, value)

            new_configuration = config.energy_assistant_config.devices.get_device_config(device_id)

            device_name = data.get("name", persisted_device.name)
            device_icon = data.get("icon", persisted_device.icon)
            device_type = data.get("type", persisted_device.type)
            await persisted_device.update(
                session, device_name, device_icon, persisted_device.power_mode, device_type, data
            )
            device.configure(new_configuration)
            return ConfigModel.model_validate({"config": new_configuration})



"""Views for home measurement API."""

import uuid
from typing import Annotated

from fastapi import APIRouter, Depends, Request

from .schema import ReadAllSessionLogEntriesResponse
from .use_cases import ReadAllLogEntries

router = APIRouter(prefix="/sessionlog")


@router.get("", response_model=ReadAllSessionLogEntriesResponse)
async def read_all(
    request: Request,
    use_case: Annotated[ReadAllLogEntries, Depends(ReadAllLogEntries)],
    device_id: uuid.UUID | None = None,
) -> ReadAllSessionLogEntriesResponse:
    """Rest end point for read all devices."""
    return ReadAllSessionLogEntriesResponse(entries=[entry async for entry in use_case.execute(device_id)])



"""Mqtt connection for energy assistant."""

import copy
import json
import logging
import pathlib
import pickle
import uuid
from datetime import UTC, datetime

import emhass  # type: ignore
import numpy as np
import pandas as pd
from emhass import utils  # type: ignore
from emhass.forecast import Forecast  # type: ignore
from emhass.machine_learning_forecaster import MLForecaster  # type: ignore
from emhass.optimization import Optimization  # type: ignore
from emhass.retrieve_hass import RetrieveHass  # type: ignore
from sklearn.metrics import r2_score  # type: ignore

from energy_assistant import Optimizer
from energy_assistant.devices import LoadInfo, Location, StateId, StatesRepository
from energy_assistant.devices.analysis import DataBuffer, create_timeseries_from_const
from energy_assistant.devices.home import Home
from energy_assistant.devices.homeassistant import HOMEASSISTANT_CHANNEL, Homeassistant
from energy_assistant.models.forecast import ForecastSchema, ForecastSerieSchema
from energy_assistant.storage.config import ConfigStorage

from .constants import ROOT_LOGGER_NAME

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)

SENSOR_POWER_NO_VAR_LOADS = "power_load_no_var_loads"
DEFAULT_HASS_ENTITY_PREFIX = "em"
DEFAULT_COST_FUNC = "profit"
LOAD_FORECAST_MODEL_TYPE = "load_forecast"


class MLForecasterTuneError(Exception):
    """Error while tuning the ML forecast."""

    def __init__(self) -> None:
        """Create an MLForecasterTuneError instance."""
        super().__init__("The ML forecaster file was not found, please run a model fit method before this tune method")


class OptimizerNotInitializedError(Exception):
    """Optimizer not initialized error."""

    def __init__(self) -> None:
        """Create an OptimizerNotInitializedError instance."""
        super().__init__("Optimizer forecast is not initialized.")


class EmhassOptimizer(Optimizer):
    """Optimizer based on Emhass."""

    def __init__(self, data_folder: str, config: ConfigStorage, hass: Homeassistant, location: Location) -> None:
        """Create an emhass optimizer instance."""
        self._data_folder: pathlib.Path = pathlib.Path(data_folder)
        self._logger = LOGGER
        self._hass = hass
        self._hass_url: str = hass.url
        if self._hass_url is not None and self._hass_url[-1] != "/":
            self._hass_url = self._hass_url + "/"
        self._hass_token: str = hass.token
        self._location: Location = location

        home_config = config.home
        self._solar_power_id: str | None = None
        if home_config is not None:
            self._solar_power_id = home_config.get("solar_power")
        self._emhass_config: dict | None = config.emhass.as_dict()

        self._emhass_path_conf = {}
        self._emhass_path_conf["data_path"] = self._data_folder
        self._emhass_path_conf["root_path"] = pathlib.Path(emhass.__file__).parent

        self._power_no_var_loads_id = f"sensor.{DEFAULT_HASS_ENTITY_PREFIX}_{SENSOR_POWER_NO_VAR_LOADS}"
        if self._emhass_config is not None:
            self._pv_forecast_method = self._emhass_config.get("pv_forecast_method", "homeassistant")
            self._cost_fun = self._emhass_config.get("costfun", DEFAULT_COST_FUNC)
            self._hass_entity_prefix = self._emhass_config.get("hass_entity_prefix", DEFAULT_HASS_ENTITY_PREFIX)
            self._power_no_var_loads_id = f"sensor.{self._hass_entity_prefix}_{SENSOR_POWER_NO_VAR_LOADS}"
            params = json.dumps(self._emhass_config)
            retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(
                self._emhass_path_conf, False, params=params
            )
            # Patch variables with Energy Assistant Config
            retrieve_hass_conf["hass_url"] = self._hass_url
            retrieve_hass_conf["long_lived_token"] = self._hass_token
            retrieve_hass_conf["var_PV"] = self._solar_power_id
            if "var_load" not in retrieve_hass_conf:
                retrieve_hass_conf["var_load"] = self._power_no_var_loads_id
            else:
                self._power_no_var_loads_id = retrieve_hass_conf["var_load"]
            retrieve_hass_conf["var_replace_zero"] = [self._solar_power_id]
            retrieve_hass_conf["var_interp"] = [
                self._solar_power_id,
                self._power_no_var_loads_id,
            ]

            retrieve_hass_conf["time_zone"] = self._location.get_time_zone()
            retrieve_hass_conf["lat"] = self._location.latitude
            retrieve_hass_conf["lon"] = self._location.longitude
            retrieve_hass_conf["alt"] = self._location.elevation
            if "continual_publish" not in retrieve_hass_conf:
                retrieve_hass_conf["continual_publish"] = False

            optim_conf["num_def_loads"] = 0
            if "compute_curtailment" not in plant_conf:
                plant_conf["compute_curtailment"] = False

            self._retrieve_hass_conf = retrieve_hass_conf
            self._optim_conf = optim_conf
            self._plant_conf = plant_conf

            self._method_ts_round = retrieve_hass_conf.get("method_ts_round")

            # Define main objects
            self._RetrieveHass = RetrieveHass(
                self._hass_url,
                self._hass_token,
                retrieve_hass_conf["freq"],
                self._location.get_time_zone(),
                params,
                self._emhass_path_conf,
                self._logger,
                get_data_from_file=False,
            )  # type: ignore

        if self._cost_fun is None:
            self._cost_fun = "profit"
        if self._method_ts_round is None:
            self._method_ts_round = "nearest"

        self._day_ahead_forecast: pd.DataFrame | None = None
        self._optimzed_devices: list = []
        self._projected_load_devices: list[LoadInfo] = []
        self._pv: DataBuffer = DataBuffer()
        self._no_var_loads: DataBuffer = DataBuffer()

    def update_repository_states(self, home: Home, state_repository: StatesRepository) -> None:
        """Calculate the power of the non variable/non controllable loads."""
        power = home.home_consumption_power
        for device in home.devices:
            if device.power_controllable:
                power = power - device.power
        if power < 0:
            power = 0.0
        self._no_var_loads.add_data_point(power)
        attributes = {
            "unit_of_measurement": "W",
            "state_class": "measurement",
            "device_class": "power",
        }
        state_repository.set_state(
            StateId(id=self._power_no_var_loads_id, channel=HOMEASSISTANT_CHANNEL),
            str(power),
            attributes,
        )
        state_repository.set_state(
            StateId(id=f"sensor.{self._hass_entity_prefix}_p_pv", channel=HOMEASSISTANT_CHANNEL),
            str(self._get_forecast_value("P_PV")),
            attributes,
        )
        state_repository.set_state(
            StateId(id=f"sensor.{self._hass_entity_prefix}_p_consumption", channel=HOMEASSISTANT_CHANNEL),
            str(self._get_forecast_value("P_Load")),
            attributes,
        )

    def perfect_forecast_optim(self, save_data_to_file: bool = True, debug: bool = False) -> pd.DataFrame:
        """Perform a call to the perfect forecast optimization routine.

        :param input_data_dict:  A dictionary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional
        :return: The output data of the optimization
        :rtype: pd.DataFrame

        """
        self._logger.info("Setting up needed data")

        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            None,
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "perfect-optim",
            self._logger,
        )  # type: ignore
        fcst = Forecast(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            self._emhass_path_conf,
            self._logger,
            get_data_from_file=False,
        )
        opt = Optimization(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            self._emhass_path_conf,
            self._logger,
        )

        days_list = utils.get_days_list(self._retrieve_hass_conf["days_to_retrieve"])
        var_list = [self._solar_power_id, self._power_no_var_loads_id]
        self._RetrieveHass.get_data(days_list, var_list, minimal_response=False, significant_changes_only=False)
        self._RetrieveHass.prepare_data(
            self._retrieve_hass_conf["var_load"],
            load_negative=self._retrieve_hass_conf["load_negative"],
            set_zero_min=self._retrieve_hass_conf["set_zero_min"],
            var_replace_zero=self._retrieve_hass_conf["var_replace_zero"],
            var_interp=self._retrieve_hass_conf["var_interp"],
        )
        df_input_data = self._RetrieveHass.df_final.copy()

        self._logger.info("Performing perfect forecast optimization")
        # Load cost and prod price forecast
        df_input_data = fcst.get_load_cost_forecast(df_input_data, method=fcst.optim_conf["load_cost_forecast_method"])
        df_input_data = fcst.get_prod_price_forecast(
            df_input_data,
            method=fcst.optim_conf["prod_price_forecast_method"],
        )
        opt_res = opt.perform_perfect_forecast_optim(df_input_data, days_list)
        # Save CSV file for analysis
        if save_data_to_file:
            filename = f"opt_res_perfect_optim_{self._cost_fun}.csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_perfect_optim_latest.csv"
        if not debug:
            opt_res.to_csv(self._data_folder / filename, index_label="timestamp")
        return opt_res

    def get_ml_runtime_params(self) -> dict:
        """Get the emhass runtime params for the machine learning load prediction."""
        freq = self._retrieve_hass_conf["freq"].total_seconds() / 3600

        runtimeparams: dict = {
            "num_def_loads": len(self._optimzed_devices),
            "P_deferrable_nom": [device.nominal_power for device in self._optimzed_devices],
            "def_total_hours": [max(round(device.duration / 3600), 1) for device in self._optimzed_devices],
            "def_start_timestep": [device.start_timestep for device in self._optimzed_devices],
            "def_end_timestep": [device.end_timestep for device in self._optimzed_devices],
            "treat_def_as_semi_cont": [not device.is_continous for device in self._optimzed_devices],
            "set_def_constant": [device.is_constant for device in self._optimzed_devices],
            "days_to_retrieve": self._retrieve_hass_conf.get("days_to_retrieve", 10),
            "model_type": LOAD_FORECAST_MODEL_TYPE,
            "var_model": self._power_no_var_loads_id,
            "sklearn_model": "KNeighborsRegressor",
            "num_lags": int(24 / freq),  # should be one day * 30 min
            "split_date_delta": "48h",
            "perform_backtest": True,
        }
        return runtimeparams

    async def async_get_pv_forecast(
        self,
        fcst: Forecast,
        set_mix_forecast: bool | None = False,
        df_now: pd.DataFrame | None = None,
    ) -> pd.Series:
        """Get the PV forecast."""
        if self._pv_forecast_method != "homeassistant":
            df_weather = fcst.get_weather_forecast(method=self._optim_conf["weather_forecast_method"])
            if df_now is None:
                df_now = pd.DataFrame()
            return fcst.get_power_from_weather(df_weather, set_mix_forecast, df_now)
        pv_forecast_hourly = await self._hass.get_solar_forecast()
        freq = self._retrieve_hass_conf["freq"]
        pv_forecast = pv_forecast_hourly.resample(freq).mean().interpolate()
        start = fcst.forecast_dates[0]
        end = fcst.forecast_dates[-1]
        pv_forecast_in_range = pv_forecast.loc[(pv_forecast.index >= start) & (pv_forecast.index <= end)]
        pv_forecast_serie = pv_forecast_in_range["sum"]
        pv_forecast_serie.index = pv_forecast_serie.index.tz_convert(self._location.get_time_zone())
        return pv_forecast_serie

    async def async_dayahead_forecast_optim(self, save_data_to_file: bool = False, debug: bool = False) -> None:
        """Perform a call to the day-ahead optimization routine.

        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional

        """
        self._logger.info("Setting up needed data for a day ahead forecast")

        # Fall back to the naive method in case the model for the mlforecaster does not exist
        filename_path = self._data_folder / "load_forecast_mlf.pkl"
        if not filename_path.exists():
            self._optim_conf["load_forecast_method"] = "naive"
            self._logger.warning("Falling back to the naive load forecaster.")

        # Treat runtimeparams
        params: str = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "dayahead-optim",
            self._logger,
        )[0]
        fcst = Forecast(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            self._emhass_path_conf,
            self._logger,
            get_data_from_file=False,
        )
        opt = Optimization(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            self._emhass_path_conf,
            self._logger,
        )

        pv_forecast = await self.async_get_pv_forecast(fcst)
        try:
            p_load_forecast = fcst.get_load_forecast(method=self._optim_conf["load_forecast_method"])
            p_load_forecast_values = np.array(p_load_forecast.values)
        except Exception:
            self._logger.warning(
                "Forecasting the load failed, probably due to missing history data in Home Assistant. ",
            )
            avg_non_var_power = self._no_var_loads.average()
            p_load_forecast = pd.Series(
                [avg_non_var_power for x in pv_forecast.values],
                index=pv_forecast.index,
            )
            p_load_forecast_values = np.array([avg_non_var_power for x in pv_forecast.values])

        freq = self._retrieve_hass_conf["freq"]

        df_input_data_dayahead = pd.DataFrame(
            np.transpose(np.vstack([np.array(pv_forecast.values), p_load_forecast_values])),
            index=pv_forecast.index,
            columns=["P_PV_forecast", "P_non_deferrable_load_forecast"],
        )

        projected_load = None
        for load_info in self._projected_load_devices:
            if not load_info.is_deferrable:
                series = create_timeseries_from_const(
                    load_info.nominal_power,
                    pd.Timedelta(int(load_info.duration), "s"),
                    freq,
                )
                projected_load = series if projected_load is None else projected_load + series

        if projected_load is not None:
            projected_load = projected_load.tz_convert(self._location.get_time_zone())
            df_input_data_dayahead["P_projected_load"] = projected_load
            df_input_data_dayahead["P_projected_load"] = df_input_data_dayahead["P_projected_load"].fillna(0)
        else:
            df_input_data_dayahead["P_projected_load"] = 0

        df_input_data_dayahead["P_load_forecast"] = (
            df_input_data_dayahead["P_projected_load"] + df_input_data_dayahead["P_non_deferrable_load_forecast"]
        )
        p_load_forecast = df_input_data_dayahead["P_load_forecast"]
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)

        # params_dayahead: dict = json.loads(params)
        # if 'prediction_horizon' in params_dayahead['passed_data'] and params_dayahead['passed_data']['prediction_horizon'] is not None:
        #    prediction_horizon = params_dayahead['passed_data']['prediction_horizon']
        #    df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[df_input_data_dayahead.index[0]:df_input_data_dayahead.index[prediction_horizon-1]]

        self._logger.info("Performing day-ahead forecast optimization")
        # Load cost and prod price forecast
        df_input_data_dayahead = fcst.get_load_cost_forecast(
            df_input_data_dayahead,
            method=fcst.optim_conf["load_cost_forecast_method"],
        )
        df_input_data_dayahead = fcst.get_prod_price_forecast(
            df_input_data_dayahead,
            method=fcst.optim_conf["prod_price_forecast_method"],
        )
        self._day_ahead_forecast = opt.perform_dayahead_forecast_optim(
            df_input_data_dayahead,
            pv_forecast,
            p_load_forecast,
        )
        if self._day_ahead_forecast is not None:
            self._day_ahead_forecast["P_projected_load"] = df_input_data_dayahead["P_projected_load"].copy()

        if not debug and self._day_ahead_forecast is not None:
            # Save CSV file for publish_data
            if save_data_to_file:
                today = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
                filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
            else:  # Just save the latest optimization results
                filename = "opt_res_latest.csv"
            self._day_ahead_forecast.to_csv(self._data_folder / filename, index_label="timestamp")

    async def async_naive_mpc_optim(self, save_data_to_file: bool = False, debug: bool = False) -> pd.DataFrame:
        """Perform a call to the naive Model Predictive Controller optimization routine.

        :param input_data_dict:  A dictionary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional
        :return: The output data of the optimization
        :rtype: pd.DataFrame

        """

        self._logger.info("Setting up needed data")

        runtimeparams: dict = {"prediction_horizon": 48}  # How many 30 min slots do we predict -> 24h

        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(runtimeparams),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "naive-mpc-optim",
            self._logger,
        )  # type: ignore
        fcst = Forecast(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            self._emhass_path_conf,
            self._logger,
            get_data_from_file=False,
        )
        opt = Optimization(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            self._emhass_path_conf,
            self._logger,
        )

        # Retrieve data from hass
        days_list = utils.get_days_list(1)
        var_list = [self._solar_power_id, self._power_no_var_loads_id]
        self._RetrieveHass.get_data(days_list, var_list, minimal_response=False, significant_changes_only=False)
        self._RetrieveHass.prepare_data(
            self._retrieve_hass_conf["var_load"],
            load_negative=self._retrieve_hass_conf["load_negative"],
            set_zero_min=self._retrieve_hass_conf["set_zero_min"],
            var_replace_zero=self._retrieve_hass_conf["var_replace_zero"],
            var_interp=self._retrieve_hass_conf["var_interp"],
        )
        df_input_data = self._RetrieveHass.df_final.copy()

        # Get PV and load forecasts
        p_pv_forecast = await self.async_get_pv_forecast(fcst, set_mix_forecast=True, df_now=df_input_data)

        p_load_forecast = fcst.get_load_forecast(
            method=self._optim_conf["load_forecast_method"],
            set_mix_forecast=True,
            df_now=df_input_data,
        )
        df_input_data_dayahead = pd.concat(
            [
                pd.Series(p_pv_forecast, name="P_PV_forecast"),
                pd.Series(p_load_forecast, name="P_load_forecast"),
            ],
            axis=1,
        )
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)

        # params_naive_mpc_optim: dict = json.loads(params)
        # if 'prediction_horizon' in params_naive_mpc_optim['passed_data'] and params_naive_mpc_optim['passed_data']['prediction_horizon'] is not None:
        #    prediction_horizon = params_naive_mpc_optim['passed_data']['prediction_horizon']
        #    df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[df_input_data_dayahead.index[0]:df_input_data_dayahead.index[prediction_horizon-1]]

        self._logger.info("Performing naive MPC optimization")
        # Load cost and prod price forecast
        df_input_data_dayahead = fcst.get_load_cost_forecast(
            df_input_data_dayahead,
            method=fcst.optim_conf["load_cost_forecast_method"],
        )
        df_input_data_dayahead = fcst.get_prod_price_forecast(
            df_input_data_dayahead,
            method=fcst.optim_conf["prod_price_forecast_method"],
        )

        # The specifics params for the MPC at runtime
        # TODO: Make this real parameters
        params_dict = json.loads(params)
        prediction_horizon = params_dict["passed_data"]["prediction_horizon"]
        soc_init = params_dict["passed_data"]["soc_init"]
        soc_final = params_dict["passed_data"]["soc_final"]
        def_total_hours = [
            1,
            1,
        ]  #  input_data_dict['params']['passed_data']['def_total_hours']

        opt_res_naive_mpc = opt.perform_naive_mpc_optim(
            df_input_data_dayahead,
            p_pv_forecast,
            p_load_forecast,
            prediction_horizon,
            soc_init,
            soc_final,
            def_total_hours,
        )
        # Save CSV file for publish_data
        if save_data_to_file:
            today = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
            filename = "opt_res_naive_mpc_" + today.strftime("%Y_%m_%d") + ".csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_naive_mpc_latest.csv"
        if not debug:
            opt_res_naive_mpc.to_csv(self._data_folder / filename, index_label="timestamp")
        return opt_res_naive_mpc

    def forecast_model_fit(
        self,
        only_if_file_does_not_exist: bool = False,
        days_to_retrieve: int | None = None,
    ) -> float:
        """Perform a forecast model fit from training data retrieved from Home Assistant."""

        filename = LOAD_FORECAST_MODEL_TYPE + "_mlf.pkl"
        filename_path = self._data_folder / filename

        if only_if_file_does_not_exist and filename_path.is_file():
            self._logger.info("Skip model creation")
            return 0

        self._logger.info("Setting up needed data")

        # Treat runtimeparams
        params: str = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "forecast-model-fit",
            self._logger,
        )[0]  # type: ignore

        params_dict: dict = json.loads(params)
        # Retrieve data from hass
        if days_to_retrieve is None:
            days_to_retrieve = self._retrieve_hass_conf.get("days_to_retrieve", 10)

        days_list = utils.get_days_list(days_to_retrieve)
        var_list = [self._power_no_var_loads_id]
        self._RetrieveHass.get_data(days_list, var_list)
        df_input_data = self._RetrieveHass.df_final.copy()

        data = copy.deepcopy(df_input_data)
        model_type = params_dict["passed_data"]["model_type"]
        sklearn_model = params_dict["passed_data"]["sklearn_model"]
        num_lags = params_dict["passed_data"]["num_lags"]
        split_date_delta = params_dict["passed_data"]["split_date_delta"]
        perform_backtest = params_dict["passed_data"]["perform_backtest"]
        # The ML forecaster object
        mlf = MLForecaster(
            data,
            model_type,
            self._power_no_var_loads_id,
            sklearn_model,
            num_lags,
            self._emhass_path_conf,
            self._logger,
        )
        # Fit the ML model
        df_pred = mlf.fit(split_date_delta=split_date_delta, perform_backtest=perform_backtest)
        predictions = df_pred["pred"].dropna()
        test_data = df_pred["test"].dropna()
        r2 = r2_score(test_data, predictions)
        self._logger.info(f"R2 score = {r2}")
        # Save model
        with (self._data_folder / filename).open("wb") as outp:
            pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
        return r2

    def forecast_model_tune(self) -> tuple[pd.DataFrame, MLForecaster]:
        """Tune a forecast model hyperparameters using bayesian optimization.

        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: bool, optional
        :return: The DataFrame containing the forecast data results using the optimized model
        :rtype: pd.DataFrame
        """
        # Load model
        self._logger.info("Tune the forecast model")

        filename = LOAD_FORECAST_MODEL_TYPE + "_mlf.pkl"
        filename_path = self._data_folder / filename
        if filename_path.is_file():
            with filename_path.open("rb") as inp:
                mlf = pickle.load(inp)
            # Tune the model
            df_pred_optim = mlf.tune(debug=False)

            # Save model
            with filename_path.open("wb") as outp:
                pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
            return df_pred_optim, mlf

        self._logger.error(
            "The ML forecaster file was not found, please run a model fit method before this tune method",
        )
        raise MLForecasterTuneError

    def forecast_model_predict(
        self,
        use_last_window: bool = True,
        debug: bool = False,
        mlf: MLForecaster | None = None,
    ) -> pd.Series | None:
        """Perform a forecast model predict using a previously trained skforecast model.

        :param input_data_dict: A dictionary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging.Logger
        :param use_last_window: True if the 'last_window' option should be used for the \
            custom machine learning forecast model. The 'last_window=True' means that the data \
            that will be used to generate the new forecast will be freshly retrieved from \
            Home Assistant. This data is needed because the forecast model is an auto-regressive \
            model with lags. If 'False' then the data using during the model train is used. Defaults to True
        :type use_last_window: Optional[bool], optional
        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: Optional[bool], optional
        :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
            and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
        :type mlf: Optional[mlforecaster], optional
        :return: The DataFrame containing the forecast prediction data
        :rtype: pd.DataFrame
        """
        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "forecast-model-fit",
            self._logger,
        )  # type: ignore

        params_dict: dict = json.loads(params)
        # Retrieve data from hass
        days_to_retrieve = params_dict["passed_data"]["days_to_retrieve"]

        days_list = utils.get_days_list(days_to_retrieve)
        var_list = [self._power_no_var_loads_id]
        self._RetrieveHass.get_data(days_list, var_list)
        df_input_data = self._RetrieveHass.df_final.copy()

        # Load model
        model_type = "load_forecast"
        filename = model_type + "_mlf.pkl"
        filename_path = self._data_folder / filename
        if not debug:
            if filename_path.is_file():
                with filename_path.open("rb") as inp:
                    mlf = pickle.load(inp)
            else:
                self._logger.error(
                    "The ML forecaster file was not found, please run a model fit method before this predict method",
                )
                return None
        # Make predictions
        data_last_window = copy.deepcopy(df_input_data) if use_last_window else None
        if mlf is not None:
            return mlf.predict(data_last_window)
        return None

    async def async_get_forecast(self) -> ForecastSchema:
        """Get the previously calculated forecast."""
        temp_folder = self._data_folder / "temp"
        forecast_filename = "forecast.csv"

        if self._day_ahead_forecast is None:
            try:
                self._day_ahead_forecast = pd.read_csv(temp_folder / forecast_filename)
            except Exception:
                self._logger.exception(f"{forecast_filename} is not available. Creating forecast...")
                await self.async_dayahead_forecast_optim()

        if self._day_ahead_forecast is not None:
            freq = self._retrieve_hass_conf["freq"]
            temp_folder.mkdir(parents=True, exist_ok=True)
            pv_df = self._pv.get_data_frame(freq, self._location.get_time_zone(), "pv", temp_folder)
            pv_df.to_csv(temp_folder / "pv_df.csv")
            no_var_load_df = self._no_var_loads.get_data_frame(
                freq,
                self._location.get_time_zone(),
                "non_var_loads",
                temp_folder,
            )
            df = self._day_ahead_forecast.merge(pv_df, how="left", left_index=True, right_index=True)
            df = df.merge(no_var_load_df, how="left", left_index=True, right_index=True)

            df = df.rename(columns={"P_PV": "pv_forecast"})
            df.to_csv(temp_folder / forecast_filename, index_label="time_stamp")

            while not pd.notnull(df["pv_forecast"][0]) and len(df.index) > 0:
                df = df.drop(df.index[0])

            pv_series = [x for x in df["pv"].to_list() if pd.notnull(x)]
            no_var_load_series = [x for x in df["non_var_loads"].to_list() if pd.notnull(x)]

            # TODO: Should be removed
            df = df.fillna(-10000)
            time_series = df.index.to_series()
            time: list[datetime] = time_series.tolist()
            pv_forecast = df["pv_forecast"].to_list()
            load = df["P_Load"].to_list()
            cost_profit = df["cost_profit"].to_list()

            series = [
                ForecastSerieSchema(name="pv_forecast", data=pv_forecast),
                ForecastSerieSchema(name="pv", data=pv_series),
                ForecastSerieSchema(name="consumption", data=load),
                ForecastSerieSchema(name="no_var_loads", data=no_var_load_series),
                ForecastSerieSchema(name="cost_profit", data=cost_profit),
            ]
            consumed_energy = sum(load)
            for i, d in enumerate(self._optimzed_devices):
                device = self._day_ahead_forecast[f"P_deferrable{i}"].to_list()
                series.append(ForecastSerieSchema(name=str(d.device_id), data=device))
                consumed_energy = consumed_energy + sum(device)

            period = freq.total_seconds() / 3600
            solar_energy = sum(pv_forecast) * period / 1000  # kWh
            consumed_energy = consumed_energy * period / 1000  # kWh
            cost = sum(cost_profit)

            return ForecastSchema(
                solar_energy=solar_energy,
                cost=cost,
                consumed_energy=consumed_energy,
                time=time,
                series=series,
            )
        raise OptimizerNotInitializedError

    def get_optimized_power(self, device_id: uuid.UUID) -> float:
        """Get the optimized power budget for a give device."""
        if self._day_ahead_forecast is not None:
            for i, deferrable_load_info in enumerate(self._optimzed_devices):
                if deferrable_load_info.device_id == device_id:
                    column_name = f"P_deferrable{i}"
                    return self._get_forecast_value(column_name)
        return -1

    def _get_forecast_value(self, column_name: str) -> float:
        """Get a forecasted value."""
        if self._day_ahead_forecast is not None:
            now_precise = datetime.now(self._location.get_time_zone()).replace(second=0, microsecond=0)
            if self._method_ts_round == "nearest":
                method = "nearest"
            elif self._method_ts_round == "first":
                method = "ffill"
            elif self._method_ts_round == "last":
                method = "bfill"
            else:
                method = "nearest"

            idx_closest = self._day_ahead_forecast.index.get_indexer([now_precise], method=method)[0]  # type: ignore
            if idx_closest == -1:
                idx_closest = self._day_ahead_forecast.index.get_indexer(  # type: ignore
                    [now_precise],
                    method="nearest",
                )[0]

            value = self._day_ahead_forecast.iloc[idx_closest][column_name]
            return float(value)
        return -1

    def _has_deferrable_load(self, device_id: uuid.UUID) -> bool:
        return any(deferrable_load_info.device_id == device_id for deferrable_load_info in self._optimzed_devices)

    async def async_update_devices(self, home: Home) -> None:
        """Update the selected devices from the list of devices."""
        new_optimizhed_devices = []
        needs_update = False
        self._pv.add_data_point(home.solar_production_power)
        self._projected_load_devices.clear()
        for device in home.devices:
            load_info = device.get_load_info()
            if load_info is not None:
                if load_info.is_deferrable:
                    if not self._has_deferrable_load(device.id):
                        needs_update = True
                    new_optimizhed_devices.append(load_info)
                else:
                    self._projected_load_devices.append(load_info)
                    # TODO: Only if not already there
                    needs_update = True
            elif self._has_deferrable_load(device.id):
                needs_update = True
        self._optimzed_devices = new_optimizhed_devices
        if needs_update or self._day_ahead_forecast is None:
            await self.async_dayahead_forecast_optim()



"""add utility meter

Revision ID: 2dbe1d6a4fa6
Revises: 81f94b4136ee
Create Date: 2023-11-02 17:25:37.856060

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "2dbe1d6a4fa6"
down_revision = "81f94b4136ee"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "utility_meter",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("last_meter_value", sa.Float(), nullable=False),
        sa.Column("device_id", sa.CHAR(32), nullable=False),
        sa.ForeignKeyConstraint(
            ["device_id"], ["devices.id"], name=op.f("fk_utility_meter_device_id_devices")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_utility_meter")),
        sa.UniqueConstraint("id", name=op.f("uq_utility_meter_id")),
    )
    with op.batch_alter_table("DeviceMeasurement") as batch_op:
        batch_op.alter_column("device_id", existing_type=sa.CHAR(length=32), nullable=False)
    with op.batch_alter_table("devices") as batch_op:
        batch_op.alter_column("power_mode", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.create_unique_constraint(op.f("uq_devices_id"), ["id"])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(op.f("uq_devices_id"), "devices", type_="unique")
    with op.batch_alter_table("devices") as batch_op:
        batch_op.drop_constraint(op.f("uq_devices_id"), type_="unique")
        batch_op.alter_column("devices", "power_mode", existing_type=sa.VARCHAR(), nullable=True)
    with op.batch_alter_table("DeviceMeasurement") as batch_op:
        batch_op.alter_column(
            "DeviceMeasurement", "device_id", existing_type=sa.CHAR(length=32), nullable=True
        )
    op.drop_table("utility_meter")
    # ### end Alembic commands ###



"""create session log table

Revision ID: b6020d5b13fc
Revises: efac279d8401
Create Date: 2023-07-24 12:58:37.970137

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "b6020d5b13fc"
down_revision = "efac279d8401"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "session_log",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("text", sa.String(), nullable=False),
        sa.Column("device_id", sa.CHAR(32), nullable=False),
        sa.Column("start", sa.DateTime(), nullable=False),
        sa.Column("end", sa.DateTime(), nullable=False),
        sa.Column("start_consumed_energy", sa.Float(), nullable=False),
        sa.Column("start_solar_consumed_energy", sa.Float(), nullable=False),
        sa.Column("end_consumed_energy", sa.Float(), nullable=False),
        sa.Column("end_solar_consumed_energy", sa.Float(), nullable=False),
        sa.ForeignKeyConstraint(
            ["device_id"], ["devices.id"], name=op.f("fk_session_log_device_id_devices")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_session_log")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("session_log")
    # ### end Alembic commands ###



"""The on / off controller switches the power of a device based on the currently produced PV."""

import uuid

from energy_assistant import Optimizer
from energy_assistant.constants import POWER_HYSTERESIS
from energy_assistant.devices import OnOffState, PowerModes, StateId, StatesRepository
from energy_assistant.devices.analysis import DataBuffer
from energy_assistant.devices.homeassistant import HOMEASSISTANT_CHANNEL


class OnOffController:
    """The on / off controller switches the power of a device based on the currently produced PV."""

    def __init__(self, device_id: uuid.UUID, output_id: str) -> None:
        """Create an on off controller instance."""
        self.switch_on_delay: float = 0
        self.switch_off_delay: float = 0
        self.nominal_power: float = 0
        self.power_mode: PowerModes = PowerModes.DEVICE_CONTROLLED
        self._device_id = device_id
        self._output_id = output_id

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: DataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        output_state = state_repository.get_state(self._output_id)
        if output_state is not None:
            state: OnOffState = OnOffState.from_str(output_state.value)
            new_state: OnOffState = state
            if self.power_mode == PowerModes.PV:
                if state == OnOffState.OFF:
                    min_power = grid_exported_power_data.get_min_for(self.switch_on_delay)
                    if min_power > self.nominal_power * (1 + POWER_HYSTERESIS):
                        new_state = OnOffState.ON
                elif state == OnOffState.OFF:
                    max_power = grid_exported_power_data.get_max_for(self.switch_off_delay)
                    if max_power > self.nominal_power * (1 - POWER_HYSTERESIS):
                        new_state = OnOffState.OFF
            elif self.power_mode == PowerModes.OPTIMIZED:
                power = optimizer.get_optimized_power(self._device_id)
                new_state = OnOffState.from_bool(power > 0)
            if state != new_state:
                state_repository.set_state(
                    StateId(
                        id=self._output_id,
                        channel=HOMEASSISTANT_CHANNEL,
                    ),
                    OnOffState.ON if new_state else OnOffState.OFF,
                )



"""Schemas for home measurement api."""

from pydantic import BaseModel

from energy_assistant.models.schema import DeviceMeasurementSchema, DeviceSchema


class ReadDeviceResponse(DeviceSchema):
    """API Response for reading a device."""


class ReadAllDevicesResponse(BaseModel):
    """API Response for reading all deviced."""

    devices: list[DeviceSchema]


class ReadDeviceMeasurementsResponse(BaseModel):
    """API Response for reading home measurements."""

    device_measurements: list[DeviceMeasurementSchema]


class UpdateDevicePowerModeRequest(BaseModel):
    """API Request for setting the power mode."""

    power_mode: str


class UpdateDevicePowerModeResponse(DeviceSchema):
    """API Response for setting the power mode."""



"""Extend Devices Fields

Revision ID: 2558c9508fcc
Revises: 2dbe1d6a4fa6
Create Date: 2023-12-16 19:18:13.051874

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "2558c9508fcc"
down_revision = "2dbe1d6a4fa6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("devices", sa.Column("type", sa.String(), nullable=True))
    op.add_column("devices", sa.Column("config", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("devices", "config")
    op.drop_column("devices", "type")
    # ### end Alembic commands ###



"""First commit.

Revision ID: 0bf7066b6d6c
Revises:
Create Date: 2023-06-12 06:03:35.453664

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "0bf7066b6d6c"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "HomeMeasurement",
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.Column("consumed_energy", sa.FLOAT(), nullable=False),
        sa.Column("solar_consumed_energy", sa.FLOAT(), nullable=False),
        sa.Column("solar_produced_energy", sa.FLOAT(), nullable=False),
        sa.Column("grid_imported_energy", sa.FLOAT(), nullable=False),
        sa.Column("grid_exported_energy", sa.FLOAT(), nullable=False),
        sa.Column("date", sa.DATE(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "DeviceMeasurement",
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.Column("consumed_energy", sa.FLOAT(), nullable=False),
        sa.Column("solar_consumed_energy", sa.FLOAT(), nullable=False),
        sa.Column("date", sa.DATE(), nullable=False),
        sa.Column("home_measurement_id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["home_measurement_id"],
            ["HomeMeasurement.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id", name="uq_DeviceMeasurement_id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("DeviceMeasurement")
    op.drop_table("HomeMeasurement")
    # ### end Alembic commands ###



"""Database for Energy Assistant."""

import logging
from collections.abc import AsyncIterator
from typing import Annotated

from fastapi import Depends
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine

from energy_assistant.models.base import Base
from energy_assistant.settings import settings

from .constants import ROOT_LOGGER_NAME

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)

async_engine = create_async_engine(
    settings.DB_URI,
    pool_pre_ping=True,
    echo=settings.ECHO_SQL,
)
AsyncSessionLocal = async_sessionmaker(
    bind=async_engine,
    autoflush=False,
    future=True,
)


async def create_all() -> None:
    """Create all tables in the database."""
    async with async_engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)


async def get_session() -> AsyncIterator[async_sessionmaker]:
    """Get the session for db transactions."""
    try:
        yield AsyncSessionLocal
    except SQLAlchemyError:
        LOGGER.exception("Error from SQL Alchemy")


AsyncSession = Annotated[async_sessionmaker, Depends(get_session)]



"""Views for home measurement API."""

import uuid
from typing import Annotated

from fastapi import APIRouter, Depends, Path, Request

from energy_assistant.models.schema import DeviceSchema

from .schema import (
    ReadAllDevicesResponse,
    ReadDeviceMeasurementsResponse,
    ReadDeviceResponse,
    UpdateDevicePowerModeRequest,
    UpdateDevicePowerModeResponse,
)
from .use_cases import (
    DeleteDevice,
    ReadAllDevices,
    ReadDevice,
    ReadDeviceMeasurements,
    UpdateDevicePowerMode,
)

router = APIRouter(prefix="/devices")


@router.get("", response_model=ReadAllDevicesResponse)
async def read_all(
    request: Request,
    filter_with_session_log_enties: bool,
    use_case: Annotated[ReadAllDevices, Depends(ReadAllDevices)],
) -> ReadAllDevicesResponse:
    """Rest end point for read all devices."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    return ReadAllDevicesResponse(
        devices=[
            device
            async for device in use_case.execute(
                energy_assistant.home if energy_assistant is not None else None,
                filter_with_session_log_enties,
            )
        ],
    )


@router.get(
    "/{device_id}",
    response_model=ReadDeviceResponse,
)
async def read(
    request: Request,
    device_id: Annotated[uuid.UUID, Path(..., description="")],
    use_case: Annotated[ReadDevice, Depends(ReadDevice)],
) -> DeviceSchema:
    """REST end point for read a device."""
    return await use_case.execute(device_id, request.app.home)


@router.get("/{device_id}/measurements")
async def read_measurements(
    request: Request,
    device_id: Annotated[uuid.UUID, Path(..., description="")],
    use_case: Annotated[ReadDeviceMeasurements, Depends(ReadDeviceMeasurements)],
) -> ReadDeviceMeasurementsResponse:
    """REST end point for read a device."""
    return ReadDeviceMeasurementsResponse(
        device_measurements=[device_measurement async for device_measurement in use_case.execute(device_id)],
    )


@router.put(
    "/{device_id}/power_mode",
    response_model=UpdateDevicePowerModeResponse,
)
async def update_power_mode(
    request: Request,
    data: UpdateDevicePowerModeRequest,
    device_id: Annotated[uuid.UUID, Path(..., description="")],
    use_case: Annotated[UpdateDevicePowerMode, Depends(UpdateDevicePowerMode)],
) -> DeviceSchema:
    """Update the power mode of a device."""
    return await use_case.execute(device_id, data.power_mode, request.app.home)


@router.delete("/{device_id}", status_code=204)
async def delete(
    request: Request,
    device_id: Annotated[uuid.UUID, Path(..., description="")],
    use_case: Annotated[DeleteDevice, Depends(DeleteDevice)],
) -> None:
    """REST end point for delete a device."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    await use_case.execute(device_id, energy_assistant.home if energy_assistant is not None else None)
